import os
import re
from typing import Union
from xml.etree import ElementTree

import fanart
import tmdbsimple
from fanart.core import Request as fanartRequest

import sickchill
from sickchill import logger, settings
from sickchill.helper.common import replace_extension, try_int
from sickchill.oldbeard import helpers

from . import helpers as metadata_helpers


class GenericMetadata(object):
    """
    Base class for all metadata providers. Default behavior is meant to mostly
    follow KODI 12+ metadata standards. Has support for:
    - show metadata file
    - episode metadata file
    - episode thumbnail
    - show fanart
    - show poster
    - show banner
    - season thumbnails (poster)
    - season thumbnails (banner)
    - season all poster
    - season all banner
    """

    def __init__(
        self,
        show_metadata=False,
        episode_metadata=False,
        fanart=False,
        poster=False,
        banner=False,
        episode_thumbnails=False,
        season_posters=False,
        season_banners=False,
        season_all_poster=False,
        season_all_banner=False,
    ):

        self.name = "Generic"

        self._ep_nfo_extension = "nfo"
        self._show_metadata_filename = "tvshow.nfo"

        self.fanart_name = "fanart.jpg"
        self.poster_name = "poster.jpg"
        self.banner_name = "banner.jpg"

        self.season_all_poster_name = "season-all-poster.jpg"
        self.season_all_banner_name = "season-all-banner.jpg"

        self.show_metadata = show_metadata
        self.episode_metadata = episode_metadata
        self.fanart = fanart
        self.poster = poster
        self.banner = banner
        self.episode_thumbnails = episode_thumbnails
        self.season_posters = season_posters
        self.season_banners = season_banners
        self.season_all_poster = season_all_poster
        self.season_all_banner = season_all_banner

    def get_config(self):
        config_list = [
            self.show_metadata,
            self.episode_metadata,
            self.fanart,
            self.poster,
            self.banner,
            self.episode_thumbnails,
            self.season_posters,
            self.season_banners,
            self.season_all_poster,
            self.season_all_banner,
        ]
        return "|".join([str(int(x)) for x in config_list])

    def get_id(self):
        return GenericMetadata.makeID(self.name)

    @staticmethod
    def makeID(name):
        name_id = re.sub(r"[+]", "plus", name)
        name_id = re.sub(r"[^\w\d_]", "_", name_id).lower()
        return name_id

    def set_config(self, config_string):
        config_list = [bool(int(x)) for x in config_string.split("|")]
        self.show_metadata = config_list[0]
        self.episode_metadata = config_list[1]
        self.fanart = config_list[2]
        self.poster = config_list[3]
        self.banner = config_list[4]
        self.episode_thumbnails = config_list[5]
        self.season_posters = config_list[6]
        self.season_banners = config_list[7]
        self.season_all_poster = config_list[8]
        self.season_all_banner = config_list[9]

    @staticmethod
    def _check_exists(location):
        if location:
            result = os.path.isfile(location)
            logger.debug("Checking if " + location + " exists: " + str(result))
            return result
        return False

    def _has_show_metadata(self, show_obj):
        return self._check_exists(self.get_show_file_path(show_obj))

    def _has_episode_metadata(self, ep_obj):
        return self._check_exists(self.get_episode_file_path(ep_obj))

    def _has_fanart(self, show_obj):
        return self._check_exists(self.get_fanart_path(show_obj))

    def _has_poster(self, show_obj):
        return self._check_exists(self.get_poster_path(show_obj))

    def _has_banner(self, show_obj):
        return self._check_exists(self.get_banner_path(show_obj))

    def _has_episode_thumb(self, ep_obj):
        return self._check_exists(self.get_episode_thumb_path(ep_obj))

    def _has_season_poster(self, show_obj, season):
        return self._check_exists(self.get_season_poster_path(show_obj, season))

    def _has_season_banner(self, show_obj, season):
        return self._check_exists(self.get_season_banner_path(show_obj, season))

    def _has_season_all_poster(self, show_obj):
        return self._check_exists(self.get_season_all_poster_path(show_obj))

    def _has_season_all_banner(self, show_obj):
        return self._check_exists(self.get_season_all_banner_path(show_obj))

    def get_show_file_path(self, show_obj):
        return os.path.join(show_obj.location, self._show_metadata_filename)

    def get_episode_file_path(self, ep_obj):
        return replace_extension(ep_obj.location, self._ep_nfo_extension)

    def get_fanart_path(self, show_obj):
        return os.path.join(show_obj.location, self.fanart_name)

    def get_poster_path(self, show_obj):
        return os.path.join(show_obj.location, self.poster_name)

    def get_banner_path(self, show_obj):
        return os.path.join(show_obj.location, self.banner_name)

    @staticmethod
    def get_episode_thumb_path(ep_obj):
        """
        Returns the path where the episode thumbnail should be stored.
        ep_obj: a TVEpisode instance for which to create the thumbnail
        """
        if os.path.isfile(ep_obj.location):

            tbn_filename = ep_obj.location.rpartition(".")

            if tbn_filename[0] == "":
                tbn_filename = ep_obj.location + "-thumb.jpg"
            else:
                tbn_filename = tbn_filename[0] + "-thumb.jpg"
        else:
            return None

        return tbn_filename

    @staticmethod
    def get_season_poster_path(show_obj, season):
        """
        Returns the full path to the file for a given season poster.

        show_obj: a TVShow instance for which to generate the path
        season: a season number to be used for the path. Note that season 0
                means specials.
        """

        # Our specials thumbnail is, well, special
        if season == 0:
            season_poster_filename = "season-specials"
        else:
            season_poster_filename = "season" + str(season).zfill(2)

        return os.path.join(show_obj.location, season_poster_filename + "-poster.jpg")

    @staticmethod
    def get_season_banner_path(show_obj, season):
        """
        Returns the full path to the file for a given season banner.

        show_obj: a TVShow instance for which to generate the path
        season: a season number to be used for the path. Note that season 0
                means specials.
        """

        # Our specials thumbnail is, well, special
        if season == 0:
            season_banner_filename = "season-specials"
        else:
            season_banner_filename = "season" + str(season).zfill(2)

        return os.path.join(show_obj.location, season_banner_filename + "-banner.jpg")

    def get_season_all_poster_path(self, show_obj):
        return os.path.join(show_obj.location, self.season_all_poster_name)

    def get_season_all_banner_path(self, show_obj):
        return os.path.join(show_obj.location, self.season_all_banner_name)

    def _show_data(self, show_obj) -> Union[ElementTree.ElementTree, None]:
        """
        This should be overridden by the implementing class. It should
        provide the content of the show metadata file.
        """
        return None

    def _ep_data(self, ep_obj) -> Union[ElementTree.ElementTree, None]:
        """
        This should be overridden by the implementing class. It should
        provide the content of the episode metadata file.
        """
        return None

    def create_show_metadata(self, show_obj):
        if self.show_metadata and show_obj and not self._has_show_metadata(show_obj):
            logger.debug("Metadata provider " + self.name + " creating show metadata for " + show_obj.name)
            return self.write_show_file(show_obj)
        return False

    def update_show_indexer_metadata(self, show_obj):
        if self.show_metadata and show_obj and self._has_show_metadata(show_obj):
            logger.debug("Metadata provider " + self.name + " updating show indexer info metadata file for " + show_obj.name)

            nfo_file_path = self.get_show_file_path(show_obj)

            try:
                with open(nfo_file_path, "rb") as xmlFileObj:
                    showXML = ElementTree.ElementTree(file=xmlFileObj)

                indexerid = showXML.find("id")

                root = showXML.getroot()
                if indexerid is not None:
                    if indexerid.text == str(show_obj.indexerid):
                        return True
                    indexerid.text = str(show_obj.indexerid)
                else:
                    ElementTree.SubElement(root, "id").text = str(show_obj.indexerid)

                # Make it purdy
                helpers.indentXML(root)

                showXML.write(nfo_file_path, encoding="UTF-8")
                helpers.chmodAsParent(nfo_file_path)

                return True
            except IOError as e:
                logger.error("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? " + str(e))

    def create_episode_metadata(self, ep_obj):
        if self.episode_metadata and ep_obj and not self._has_episode_metadata(ep_obj):
            logger.debug("Metadata provider " + self.name + " creating episode metadata for " + ep_obj.pretty_name())
            return self.write_ep_file(ep_obj)
        return False

    def update_episode_metadata(self, ep_obj):
        if self.episode_metadata and ep_obj and self._has_episode_metadata(ep_obj):
            logger.debug("Metadata provider " + self.name + " updating episode indexer info metadata file for " + ep_obj.pretty_name())
            nfo_file_path = self.get_episode_file_path(ep_obj)

            attribute_map = {
                "title": "name",
                "aired": "airdate",
                "season": "season",
                "episode": "episode",
                "showtitle": "show.name",
                "runtime": "show.runtime",
                "plot": "description",
            }
            try:
                with open(nfo_file_path, "rb") as xmlFileObj:
                    episodeXML = ElementTree.ElementTree(file=xmlFileObj)

                changed = False
                for attribute in attribute_map:
                    try:
                        if not hasattr(ep_obj, attribute_map[attribute]):
                            continue

                        node = episodeXML.find(attribute)
                        if node is None or node.text == str(getattr(ep_obj, attribute_map[attribute])):
                            continue

                        node.text = str(getattr(ep_obj, attribute_map[attribute]))
                        changed = True
                    except AttributeError:
                        pass

                if not changed:
                    return True

                root = episodeXML.getroot()

                # Make it purdy
                helpers.indentXML(root)

                episodeXML.write(nfo_file_path, encoding="UTF-8")
                helpers.chmodAsParent(nfo_file_path)

                return True
            except IOError as error:
                logger.warning("Unable to write file to {} - are you sure the folder is writable? {}".format(nfo_file_path, str(error)))
            except ElementTree.ParseError as error:
                logger.warning("Error parsing existing nfo file at {} - {}".format(nfo_file_path, str(error)))

    def create_fanart(self, show_obj):
        if self.fanart and show_obj and not self._has_fanart(show_obj):
            logger.debug("Metadata provider " + self.name + " creating fanart for " + show_obj.name)
            return self.save_fanart(show_obj)
        return False

    def create_poster(self, show_obj):
        if self.poster and show_obj and not self._has_poster(show_obj):
            logger.debug("Metadata provider " + self.name + " creating poster for " + show_obj.name)
            return self.save_poster(show_obj)
        return False

    def create_banner(self, show_obj):
        if self.banner and show_obj and not self._has_banner(show_obj):
            logger.debug("Metadata provider " + self.name + " creating banner for " + show_obj.name)
            return self.save_banner(show_obj)
        return False

    def create_episode_thumb(self, ep_obj):
        if self.episode_thumbnails and ep_obj and not self._has_episode_thumb(ep_obj):
            logger.debug("Metadata provider " + self.name + " creating episode thumbnail for " + ep_obj.pretty_name())
            return self.save_thumbnail(ep_obj)
        return False

    def create_season_posters(self, show_obj):
        if self.season_posters and show_obj:
            result = []
            for season in show_obj.episodes:
                if not self._has_season_poster(show_obj, season):
                    logger.debug("Metadata provider " + self.name + " creating season posters for " + show_obj.name)
                    result.extend([self.save_season_poster(show_obj, season)])
            return all(result)
        return False

    def create_season_banners(self, show_obj):
        if self.season_banners and show_obj:
            result = []
            logger.debug("Metadata provider " + self.name + " creating season banners for " + show_obj.name)
            for season in show_obj.episodes:
                if not self._has_season_banner(show_obj, season):
                    result.extend([self.save_season_banner(show_obj, season)])
            return all(result)
        return False

    def create_season_all_poster(self, show_obj):
        if self.season_all_poster and show_obj and not self._has_season_all_poster(show_obj):
            logger.debug("Metadata provider " + self.name + " creating season all poster for " + show_obj.name)
            return self.save_season_all_poster(show_obj)
        return False

    def create_season_all_banner(self, show_obj):
        if self.season_all_banner and show_obj and not self._has_season_all_banner(show_obj):
            logger.debug("Metadata provider " + self.name + " creating season all banner for " + show_obj.name)
            return self.save_season_all_banner(show_obj)
        return False

    def write_show_file(self, show_obj):
        """
        Generates and writes show_obj's metadata under the given path to the
        filename given by get_show_file_path()

        show_obj: TVShow object for which to create the metadata

        path: An absolute or relative path where we should put the file. Note that
                the file name will be the default show_filename.

        Note that this method expects that _show_data will return an ElementTree
        object. If your _show_data returns data in another format yo'll need to
        override this method.
        """

        data = self._show_data(show_obj)

        if not data:
            return False

        nfo_file_path = self.get_show_file_path(show_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                logger.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                helpers.chmodAsParent(nfo_file_dir)

            logger.debug("Writing show nfo file to " + nfo_file_path)

            nfo_file = open(nfo_file_path, "wb")
            data.write(nfo_file, encoding="UTF-8")
            nfo_file.close()
            helpers.chmodAsParent(nfo_file_path)
        except IOError as e:
            logger.error("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? " + str(e))
            return False

        return True

    def write_ep_file(self, ep_obj):
        """
        Generates and writes ep_obj's metadata under the given path with the
        given filename root. Uses the episode's name with the extension in
        _ep_nfo_extension.

        ep_obj: TVEpisode object for which to create the metadata

        filename_path: The file name to use for this metadata. Note that the extension
                will be automatically added based on _ep_nfo_extension. This should
                include an absolute path.

        Note that this method expects that _ep_data will return an ElementTree
        object. If your _ep_data returns data in another format yo'll need to
        override this method.
        """

        data = self._ep_data(ep_obj)
        if not data:
            return False

        # def print_data(d):
        #     for child in d.getroot():
        #         print(str(child.tag), str(child.text))

        # print_data(data)

        nfo_file_path = self.get_episode_file_path(ep_obj)
        nfo_file_dir = os.path.dirname(nfo_file_path)

        try:
            if not os.path.isdir(nfo_file_dir):
                logger.debug("Metadata dir didn't exist, creating it at " + nfo_file_dir)
                os.makedirs(nfo_file_dir)
                helpers.chmodAsParent(nfo_file_dir)

            logger.debug("Writing episode nfo file to " + nfo_file_path)
            nfo_file = open(nfo_file_path, "wb")
            data.write(nfo_file, encoding="UTF-8")
            nfo_file.close()
            helpers.chmodAsParent(nfo_file_path)
        except IOError as e:
            logger.error("Unable to write file to " + nfo_file_path + " - are you sure the folder is writable? " + str(e))
            return False

        return True

    def save_thumbnail(self, ep_obj):
        """
        Retrieves a thumbnail and saves it to the correct spot. This method should not need to
        be overridden by implementing classes, changing get_episode_thumb_path and
        _get_episode_thumb_url should suffice.

        ep_obj: a TVEpisode object for which to generate a thumbnail
        """

        thumb_url = sickchill.indexer.episode_image_url(ep_obj)
        if not thumb_url:
            logger.debug("No thumb is available for this episode, not creating a thumb")
            return False

        file_path = self.get_episode_thumb_path(ep_obj)
        if not file_path:
            logger.debug("Unable to find a file path to use for this thumbnail, not generating it")
            return False

        thumb_data = metadata_helpers.getShowImage(thumb_url)
        if not thumb_data:
            logger.debug("No thumb is available for this episode, not creating a thumb")
            return False

        result = self._write_image(thumb_data, file_path)

        if not result:
            return False

        for cur_ep in [ep_obj] + ep_obj.relatedEps:
            cur_ep.hastbn = True

        return True

    def save_fanart(self, show_obj):
        """
        Downloads a fanart image and saves it to the filename specified by fanart_name
        inside the show's root folder.

        show_obj: a TVShow object for which to download fanart
        """

        # use the default fanart name
        fanart_path = self.get_fanart_path(show_obj)
        if not fanart_path:
            logger.debug("Fanart path for show {} came back blank, skipping this image".format(show_obj.name))
            return False

        fanart_url = sickchill.indexer.series_fanart_url(show_obj)
        if not fanart_url:
            fanart_url = self._retrieve_show_image_urls_from_fanart(show_obj, "fanart")
            if not fanart_url:
                logger.debug("Fanart url not found for show {}, skipping this image".format(show_obj.name))
                return False

        fanart_data = metadata_helpers.getShowImage(fanart_url)
        if not fanart_data:
            logger.debug("No fanart image was retrieved, unable to write fanart")
            return False

        return self._write_image(fanart_data, fanart_path)

    def save_poster(self, show_obj):
        """
        Downloads a poster image and saves it to the filename specified by poster_name
        inside the show's root folder.

        show_obj: a TVShow object for which to download a poster
        """

        # use the default poster name
        poster_path = self.get_poster_path(show_obj)
        if not poster_path:
            logger.debug("Banner path for show {} came back blank, skipping this image".format(show_obj.name))
            return False

        poster_url = sickchill.indexer.series_poster_url(show_obj)
        if not poster_url:
            poster_url = self._retrieve_show_image_urls_from_fanart(show_obj, "poster")
            if not poster_url:
                logger.debug("Poster url not found for show {}, skipping this image".format(show_obj.name))
                return False

        poster_data = metadata_helpers.getShowImage(poster_url)
        if not poster_data:
            logger.debug("No show poster image was retrieved, unable to write poster")
            return False

        return self._write_image(poster_data, poster_path)

    def save_banner(self, show_obj):
        """
        Downloads a banner image and saves it to the filename specified by banner_name
        inside the show's root folder.

        show_obj: a TVShow object for which to download a banner
        """

        banner_path = self.get_banner_path(show_obj)
        if not banner_path:
            logger.debug("Banner path for show {} came back blank, skipping this image".format(show_obj.name))
            return False

        banner_url = sickchill.indexer.series_banner_url(show_obj)
        if not banner_url:
            banner_url = self._retrieve_show_image_urls_from_fanart(show_obj, "banner")
            if not banner_url:
                logger.debug("Banner url not found for show {}, skipping this image".format(show_obj.name))
                return False

        banner_data = metadata_helpers.getShowImage(banner_url)
        if not banner_data:
            logger.debug("No show banner image was retrieved, unable to write banner")
            return False

        return self._write_image(banner_data, banner_path)

    def save_season_poster(self, show_obj, season):
        """
        Saves a specific season poster to disk for the given show.

        show_obj: a TVShow object for which to save the season thumbs
        """

        season_poster_url = sickchill.indexer.season_poster_url(show_obj, season)
        if not season_poster_url:
            season_poster_url = self._retrieve_show_image_urls_from_fanart(show_obj, "season_poster", season=season)
            if not season_poster_url:
                logger.debug("Season poster url not found for season {}, skipping this season".format(season))
                return False

        season_poster_file_path = self.get_season_poster_path(show_obj, season)
        if not season_poster_file_path:
            logger.debug("Path for season {} came back blank, skipping this season".format(season))
            return False

        image_data = metadata_helpers.getShowImage(season_poster_url)
        if not image_data:
            logger.debug("No season poster data available, skipping this season")
            return False

        return self._write_image(image_data, season_poster_file_path)

    def save_season_banner(self, show_obj, season):
        """
        Saves the first season banner for a season to disk for the given show.

        show_obj: a TVShow object for which to save the season thumbs
        """
        season_banner_url = sickchill.indexer.season_banner_url(show_obj, season)
        if not season_banner_url:
            season_banner_url = self._retrieve_show_image_urls_from_fanart(show_obj, "season_banner", season=season)
            if not season_banner_url:
                logger.debug("Url for season banner {} came back blank, skipping this season".format(season))
                return False

        season_banner_file_path = self.get_season_banner_path(show_obj, season)
        if not season_banner_file_path:
            logger.debug("Path for season {} came back blank, skipping this season".format(season))
            return False

        image_data = metadata_helpers.getShowImage(season_banner_url)
        if not image_data:
            logger.debug("No season banner data available, skipping this season")
            return False

        return self._write_image(image_data, season_banner_file_path)

    def save_season_all_poster(self, show_obj):
        poster_url = sickchill.indexer.series_poster_url(show_obj)
        if not poster_url:
            poster_url = self._retrieve_show_image_urls_from_fanart(show_obj, "season_poster", season=0)
            if not poster_url:
                logger.debug("Url for season all poster came back blank, skipping this season")
                return False

        season_poster_file_path = self.get_season_all_poster_path(show_obj)
        if not season_poster_file_path:
            logger.debug("Path for season all poster came back blank, skipping this season")
            return False

        image_data = metadata_helpers.getShowImage(poster_url)
        if not image_data:
            logger.debug("No season all poster data available, skipping this season")
            return False

        return self._write_image(image_data, season_poster_file_path)

    def save_season_all_banner(self, show_obj):
        banner_url = sickchill.indexer.series_banner_url(show_obj)
        if not banner_url:
            banner_url = self._retrieve_show_image_urls_from_fanart(show_obj, "season_banner", season=0)
            if not banner_url:
                logger.debug("Url for season all banner came back blank, skipping this season")
                return False

        season_banner_file_path = self.get_season_all_banner_path(show_obj)
        if not season_banner_file_path:
            logger.debug("Path for season all banner came back blank, skipping this season")
            return False

        image_data = metadata_helpers.getShowImage(banner_url)
        if not image_data:
            logger.debug("No season all banner data available, skipping this season")
            return False

        return self._write_image(image_data, season_banner_file_path)

    @staticmethod
    def _write_image(image_data, image_path, overwrite=False):
        """
        Saves the data in image_data to the location image_path. Returns True/False
        to represent success or failure.

        image_data: binary image data to write to file
        image_path: file location to save the image to
        """

        # don't bother overwriting it
        if not overwrite and os.path.isfile(image_path):
            logger.debug("Image already exists, not downloading")
            return False

        image_dir = os.path.dirname(image_path)

        if not image_data:
            logger.debug("Unable to retrieve image to save in {0}, skipping".format(image_path))
            return False

        try:
            if not os.path.isdir(image_dir):
                logger.debug("Metadata dir didn't exist, creating it at " + image_dir)
                os.makedirs(image_dir)
                helpers.chmodAsParent(image_dir)

            outFile = open(image_path, "wb")
            outFile.write(image_data)
            outFile.close()
            helpers.chmodAsParent(image_path)
        except IOError as e:
            logger.error("Unable to write image to " + image_path + " - are you sure the show folder is writable? " + str(e))
            return False

        return True

    def retrieveShowMetadata(self, folder):
        """
        Used only when mass adding Existing Shows, using previously generated Show metadata to reduce the need to query TVDB.
        """

        empty_return = (None, None, None)

        metadata_path = os.path.join(folder, self._show_metadata_filename)

        if not os.path.isdir(folder) or not os.path.isfile(metadata_path):
            logger.debug(_("Can't load the metadata file from {0}, it doesn't exist").format(metadata_path))
            return empty_return

        logger.debug(_("Loading show info from metadata file in {0}").format(metadata_path))

        def read_xml():
            with open(metadata_path, "rb") as __xml_file:
                try:
                    __show_xml = ElementTree.ElementTree(file=__xml_file)
                except (ElementTree.ParseError, IOError):
                    __show_xml = None
            return __show_xml

        def fix_xml():
            logger.info(
                _(
                    "There was an error loading {0}, trying to repair it by fixing & symbols. If it still has problems, please check the file " "manually"
                ).format(metadata_path)
            )
            with open(metadata_path) as __xml_file:
                output = __xml_file.read()

            regex = re.compile(r"&(?!amp;|lt;|gt;)")
            output = regex.sub("&amp;", output)
            with open(metadata_path, "wb") as __xml_file:
                __xml_file.write(output)

            return True

        try:
            show_xml = read_xml() or fix_xml() and read_xml()
            if not show_xml:
                logger.debug(_("Can't load the metadata file from {0}, error reading file").format(metadata_path))
                return empty_return

            if not (show_xml.findtext("title") or (show_xml.findtext("tvdbid") and show_xml.findtext("id"))):
                logger.info(
                    _("Invalid info in tvshow.nfo (missing name or id): {0} {1} {2}").format(
                        show_xml.findtext("title"), show_xml.findtext("tvdbid"), show_xml.findtext("id")
                    )
                )
                return empty_return

            name = show_xml.findtext("title")

            indexer_id_text = show_xml.findtext("tvdbid") or show_xml.findtext("id")
            if indexer_id_text:
                indexer_id = try_int(indexer_id_text, None)
                if indexer_id is None or indexer_id < 1:
                    logger.debug(_("Invalid Indexer ID ({0}), not using metadata file").format(str(indexer_id)))
                    return empty_return
            else:
                logger.debug(_("Empty <id> or <tvdbid> field in NFO, unable to find a ID, not using metadata file"))
                return empty_return

            indexer = 1
            epg_url_text = show_xml.findtext("episodeguide/url")
            if epg_url_text:
                epg_url = epg_url_text.lower()
                if str(indexer_id) in epg_url and "tvrage" in epg_url:
                    if sickchill.indexer.TVRAGE not in sickchill.indexer:
                        logger.warning(_("Invalid Indexer ID ({0}), not using metadata file because it has TVRage info").format(indexer_id))
                        return empty_return
                    return indexer_id, show_xml.findtext("title"), sickchill.indexer.TVRAGE
                if str(indexer_id) in epg_url and "tvdb" in epg_url:
                    return indexer_id, show_xml.findtext("title"), sickchill.indexer.TVDB

        except Exception as e:
            logger.warning(_("There was an error parsing your existing metadata file: '{0}' error: {1}").format(metadata_path, str(e)))
            return empty_return

        return indexer_id, name, indexer

    @staticmethod
    def _retrieve_show_image_urls_from_tmdb(show, img_type, multiple=False):
        types = {"poster": "posters", "banner": None, "fanart": "backdrops", "poster_thumb": "posters", "banner_thumb": None}

        if not types[img_type]:
            return [] if multiple else ""

        # get TMDB configuration info
        tmdbsimple.API_KEY = settings.TMDB_API_KEY
        config = tmdbsimple.Configuration()
        response = config.info()
        base_url = response["images"]["base_url"]
        sizes = response["images"]["poster_sizes"]

        def size_str_to_int(x):
            return float("inf") if x == "original" else int(x[1:])

        max_size = max(sizes, key=size_str_to_int)

        try:
            results = []
            find = tmdbsimple.Find(show.indexerid)
            found = find.info(external_source="tvdb_id")
            if found["tv_results"]:
                tmdb_show = tmdbsimple.TV(found["tv_results"][0]["id"])
                images = tmdb_show.images()
                if types[img_type] in images:
                    for result in images[types[img_type]]:
                        results.append("{0}{1}{2}".format(base_url, max_size, result["file_path"]))
                        if not multiple:
                            return results[0]
                    return results
        except Exception as error:
            logger.debug(error)

        logger.info("Could not find any " + img_type + " images on TMDB for " + show.name)

    @staticmethod
    def _retrieve_show_image_urls_from_fanart(show, img_type, thumb=False, season=None, multiple=False):
        types = {
            "poster": fanart.TYPE.TV.POSTER,
            "banner": fanart.TYPE.TV.BANNER,
            "poster_thumb": fanart.TYPE.TV.POSTER,
            "banner_thumb": fanart.TYPE.TV.BANNER,
            "fanart": fanart.TYPE.TV.BACKGROUND,
            "season_poster": fanart.TYPE.TV.SEASONPOSTER,
            "season_banner": fanart.TYPE.TV.SEASONBANNER,
        }

        try:
            if img_type in types:
                request = fanartRequest(
                    apikey=settings.FANART_API_KEY,
                    id=show.indexerid,
                    ws=fanart.WS.TV,
                    type=types[img_type],
                    sort=fanart.SORT.POPULAR,
                    limit=(fanart.LIMIT.ONE, fanart.LIMIT.ALL)[season is not None],
                )

                resp = request.response()
                results = resp[types[img_type]]
                if season:
                    results = [x for x in results if try_int(x["season"], default_value=None) == season]

                def _to_preview_url(url):
                    return re.sub("/fanart/", "/preview/", url)

                if multiple:
                    urls = [result["url"] for result in results]
                    if thumb:
                        urls = [_to_preview_url(url) for url in urls]
                    return urls
                else:
                    url = results[0]["url"]
                    if thumb:
                        url = _to_preview_url(url)
                    return url
        except Exception as error:
            logger.debug(error)

        logger.info("Could not find any " + img_type + " images on Fanart.tv for " + show.name)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["justo ahora", "en un rato"],
    ["hace %s segundos", "en %s segundos"],
    ["hace 1 minuto", "en 1 minuto"],
    ["hace %s minutos", "en %s minutos"],
    ["hace 1 hora", "en 1 hora"],
    ["hace %s horas", "en %s horas"],
    ["hace 1 día", "en 1 día"],
    ["hace %s días", "en %s días"],
    ["hace 1 semana", "en 1 semana"],
    ["hace %s semanas", "en %s semanas"],
    ["hace 1 mes", "en 1 mes"],
    ["hace %s meses", "en %s meses"],
    ["hace 1 año", "en 1 año"],
    ["hace %s años", "en %s años"]
]
        


"""
Tests for SickChill helpers
"""

import unittest

from .test_common import CommonTests
from .test_quality import QualityTests

if __name__ == "__main__":
    print('=====> Running all test in "sickchill_tests.helper" <=====')

    TEST_CLASSES = [
        CommonTests,
        QualityTests,
    ]

    for test_class in TEST_CLASSES:
        SUITE = unittest.TestLoader().loadTestsFromTestCase(test_class)
        unittest.TextTestRunner(verbosity=2).run(SUITE)



# encoding: utf-8
from __future__ import unicode_literals

import mimetypes
import os
import re
import sys
from tempfile import NamedTemporaryFile
from unicodedata import normalize

try:
    from urllib.parse import urlparse
except ImportError:
    from urlparse import urlparse

import requests
from twitter import TwitterError
import twitter

if sys.version_info < (3,):
    range = xrange

if sys.version_info > (3,):
    unicode = str

CHAR_RANGES = [
    range(0, 4351),
    range(8192, 8205),
    range(8208, 8223),
    range(8242, 8247)]

TLDS = [
    "ac", "ad", "ae", "af", "ag", "ai", "al", "am", "an", "ao", "aq", "ar",
    "as", "at", "au", "aw", "ax", "az", "ba", "bb", "bd", "be", "bf", "bg",
    "bh", "bi", "bj", "bl", "bm", "bn", "bo", "bq", "br", "bs", "bt", "bv",
    "bw", "by", "bz", "ca", "cc", "cd", "cf", "cg", "ch", "ci", "ck", "cl",
    "cm", "cn", "co", "cr", "cu", "cv", "cw", "cx", "cy", "cz", "de", "dj",
    "dk", "dm", "do", "dz", "ec", "ee", "eg", "eh", "er", "es", "et", "eu",
    "fi", "fj", "fk", "fm", "fo", "fr", "ga", "gb", "gd", "ge", "gf", "gg",
    "gh", "gi", "gl", "gm", "gn", "gp", "gq", "gr", "gs", "gt", "gu", "gw",
    "gy", "hk", "hm", "hn", "hr", "ht", "hu", "id", "ie", "il", "im", "in",
    "io", "iq", "ir", "is", "it", "je", "jm", "jo", "jp", "ke", "kg", "kh",
    "ki", "km", "kn", "kp", "kr", "kw", "ky", "kz", "la", "lb", "lc", "li",
    "lk", "lr", "ls", "lt", "lu", "lv", "ly", "ma", "mc", "md", "me", "mf",
    "mg", "mh", "mk", "ml", "mm", "mn", "mo", "mp", "mq", "mr", "ms", "mt",
    "mu", "mv", "mw", "mx", "my", "mz", "na", "nc", "ne", "nf", "ng", "ni",
    "nl", "no", "np", "nr", "nu", "nz", "om", "pa", "pe", "pf", "pg", "ph",
    "pk", "pl", "pm", "pn", "pr", "ps", "pt", "pw", "py", "qa", "re", "ro",
    "rs", "ru", "rw", "sa", "sb", "sc", "sd", "se", "sg", "sh", "si", "sj",
    "sk", "sl", "sm", "sn", "so", "sr", "ss", "st", "su", "sv", "sx", "sy",
    "sz", "tc", "td", "tf", "tg", "th", "tj", "tk", "tl", "tm", "tn", "to",
    "tp", "tr", "tt", "tv", "tw", "tz", "ua", "ug", "uk", "um", "us", "uy",
    "uz", "va", "vc", "ve", "vg", "vi", "vn", "vu", "wf", "ws", "ye", "yt",
    "za", "zm", "zw", "ελ", "бел", "мкд", "мон", "рф", "срб", "укр", "қаз",
    "հայ", "الاردن", "الجزائر", "السعودية", "المغرب", "امارات", "ایران", "بھارت",
    "تونس", "سودان", "سورية", "عراق", "عمان", "فلسطين", "قطر", "مصر",
    "مليسيا", "پاکستان", "भारत", "বাংলা", "ভারত", "ਭਾਰਤ", "ભારત",
    "இந்தியா", "இலங்கை", "சிங்கப்பூர்", "భారత్", "ලංකා", "ไทย",
    "გე", "中国", "中國", "台湾", "台灣", "新加坡", "澳門", "香港", "한국", "neric:",
    "abb", "abbott", "abogado", "academy", "accenture", "accountant",
    "accountants", "aco", "active", "actor", "ads", "adult", "aeg", "aero",
    "afl", "agency", "aig", "airforce", "airtel", "allfinanz", "alsace",
    "amsterdam", "android", "apartments", "app", "aquarelle", "archi", "army",
    "arpa", "asia", "associates", "attorney", "auction", "audio", "auto",
    "autos", "axa", "azure", "band", "bank", "bar", "barcelona", "barclaycard",
    "barclays", "bargains", "bauhaus", "bayern", "bbc", "bbva", "bcn", "beer",
    "bentley", "berlin", "best", "bet", "bharti", "bible", "bid", "bike",
    "bing", "bingo", "bio", "biz", "black", "blackfriday", "bloomberg", "blue",
    "bmw", "bnl", "bnpparibas", "boats", "bond", "boo", "boots", "boutique",
    "bradesco", "bridgestone", "broker", "brother", "brussels", "budapest",
    "build", "builders", "business", "buzz", "bzh", "cab", "cafe", "cal",
    "camera", "camp", "cancerresearch", "canon", "capetown", "capital",
    "caravan", "cards", "care", "career", "careers", "cars", "cartier",
    "casa", "cash", "casino", "cat", "catering", "cba", "cbn", "ceb", "center",
    "ceo", "cern", "cfa", "cfd", "chanel", "channel", "chat", "cheap",
    "chloe", "christmas", "chrome", "church", "cisco", "citic", "city",
    "claims", "cleaning", "click", "clinic", "clothing", "cloud", "club",
    "coach", "codes", "coffee", "college", "cologne", "com", "commbank",
    "community", "company", "computer", "condos", "construction", "consulting",
    "contractors", "cooking", "cool", "coop", "corsica", "country", "coupons",
    "courses", "credit", "creditcard", "cricket", "crown", "crs", "cruises",
    "cuisinella", "cymru", "cyou", "dabur", "dad", "dance", "date", "dating",
    "datsun", "day", "dclk", "deals", "degree", "delivery", "delta",
    "democrat", "dental", "dentist", "desi", "design", "dev", "diamonds",
    "diet", "digital", "direct", "directory", "discount", "dnp", "docs",
    "dog", "doha", "domains", "doosan", "download", "drive", "durban", "dvag",
    "earth", "eat", "edu", "education", "email", "emerck", "energy",
    "engineer", "engineering", "enterprises", "epson", "equipment", "erni",
    "esq", "estate", "eurovision", "eus", "events", "everbank", "exchange",
    "expert", "exposed", "express", "fage", "fail", "faith", "family", "fan",
    "fans", "farm", "fashion", "feedback", "film", "finance", "financial",
    "firmdale", "fish", "fishing", "fit", "fitness", "flights", "florist",
    "flowers", "flsmidth", "fly", "foo", "football", "forex", "forsale",
    "forum", "foundation", "frl", "frogans", "fund", "furniture", "futbol",
    "fyi", "gal", "gallery", "game", "garden", "gbiz", "gdn", "gent",
    "genting", "ggee", "gift", "gifts", "gives", "giving", "glass", "gle",
    "global", "globo", "gmail", "gmo", "gmx", "gold", "goldpoint", "golf",
    "goo", "goog", "google", "gop", "gov", "graphics", "gratis", "green",
    "gripe", "group", "guge", "guide", "guitars", "guru", "hamburg", "hangout",
    "haus", "healthcare", "help", "here", "hermes", "hiphop", "hitachi", "hiv",
    "hockey", "holdings", "holiday", "homedepot", "homes", "honda", "horse",
    "host", "hosting", "hoteles", "hotmail", "house", "how", "hsbc", "ibm",
    "icbc", "ice", "icu", "ifm", "iinet", "immo", "immobilien", "industries",
    "infiniti", "info", "ing", "ink", "institute", "insure", "int",
    "international", "investments", "ipiranga", "irish", "ist", "istanbul",
    "itau", "iwc", "java", "jcb", "jetzt", "jewelry", "jlc", "jll", "jobs",
    "joburg", "jprs", "juegos", "kaufen", "kddi", "kim", "kitchen", "kiwi",
    "koeln", "komatsu", "krd", "kred", "kyoto", "lacaixa", "lancaster", "land",
    "lasalle", "lat", "latrobe", "law", "lawyer", "lds", "lease", "leclerc",
    "legal", "lexus", "lgbt", "liaison", "lidl", "life", "lighting", "limited",
    "limo", "link", "live", "lixil", "loan", "loans", "lol", "london", "lotte",
    "lotto", "love", "ltda", "lupin", "luxe", "luxury", "madrid", "maif",
    "maison", "man", "management", "mango", "market", "marketing", "markets",
    "marriott", "mba", "media", "meet", "melbourne", "meme", "memorial", "men",
    "menu", "miami", "microsoft", "mil", "mini", "mma", "mobi", "moda", "moe",
    "mom", "monash", "money", "montblanc", "mormon", "mortgage", "moscow",
    "motorcycles", "mov", "movie", "movistar", "mtn", "mtpc", "museum",
    "nadex", "nagoya", "name", "navy", "nec", "net", "netbank", "network",
    "neustar", "new", "news", "nexus", "ngo", "nhk", "nico", "ninja", "nissan",
    "nokia", "nra", "nrw", "ntt", "nyc", "office", "okinawa", "omega", "one",
    "ong", "onl", "online", "ooo", "oracle", "orange", "org", "organic",
    "osaka", "otsuka", "ovh", "page", "panerai", "paris", "partners", "parts",
    "party", "pet", "pharmacy", "philips", "photo", "photography", "photos",
    "physio", "piaget", "pics", "pictet", "pictures", "pink", "pizza", "place",
    "play", "plumbing", "plus", "pohl", "poker", "porn", "post", "praxi",
    "press", "pro", "prod", "productions", "prof", "properties", "property",
    "pub", "qpon", "quebec", "racing", "realtor", "realty", "recipes", "red",
    "redstone", "rehab", "reise", "reisen", "reit", "ren", "rent", "rentals",
    "repair", "report", "republican", "rest", "restaurant", "review",
    "reviews", "rich", "ricoh", "rio", "rip", "rocks", "rodeo", "rsvp", "ruhr",
    "run", "ryukyu", "saarland", "sakura", "sale", "samsung", "sandvik",
    "sandvikcoromant", "sanofi", "sap", "sarl", "saxo", "sca", "scb",
    "schmidt", "scholarships", "school", "schule", "schwarz", "science",
    "scor", "scot", "seat", "seek", "sener", "services", "sew", "sex", "sexy",
    "shiksha", "shoes", "show", "shriram", "singles", "site", "ski", "sky",
    "skype", "sncf", "soccer", "social", "software", "sohu", "solar",
    "solutions", "sony", "soy", "space", "spiegel", "spreadbetting", "srl",
    "starhub", "statoil", "studio", "study", "style", "sucks", "supplies",
    "supply", "support", "surf", "surgery", "suzuki", "swatch", "swiss",
    "sydney", "systems", "taipei", "tatamotors", "tatar", "tattoo", "tax",
    "taxi", "team", "tech", "technology", "tel", "telefonica", "temasek",
    "tennis", "thd", "theater", "tickets", "tienda", "tips", "tires", "tirol",
    "today", "tokyo", "tools", "top", "toray", "toshiba", "tours", "town",
    "toyota", "toys", "trade", "trading", "training", "travel", "trust", "tui",
    "ubs", "university", "uno", "uol", "vacations", "vegas", "ventures",
    "vermögensberater", "vermögensberatung", "versicherung", "vet", "viajes",
    "video", "villas", "vin", "vision", "vista", "vistaprint", "vlaanderen",
    "vodka", "vote", "voting", "voto", "voyage", "wales", "walter", "wang",
    "watch", "webcam", "website", "wed", "wedding", "weir", "whoswho", "wien",
    "wiki", "williamhill", "win", "windows", "wine", "wme", "work", "works",
    "world", "wtc", "wtf", "xbox", "xerox", "xin", "xperia", "xxx", "xyz",
    "yachts", "yandex", "yodobashi", "yoga", "yokohama", "youtube", "zip",
    "zone", "zuerich", "дети", "ком", "москва", "онлайн", "орг", "рус", "сайт",
    "קום", "بازار", "شبكة", "كوم", "موقع", "कॉम", "नेट", "संगठन", "คอม",
    "みんな", "グーグル", "コム", "世界", "中信", "中文网", "企业", "佛山", "信息",
    "健康", "八卦", "公司", "公益", "商城", "商店", "商标", "在线", "大拿", "娱乐",
    "工行", "广东", "慈善", "我爱你", "手机", "政务", "政府", "新闻", "时尚", "机构",
    "淡马锡", "游戏", "点看", "移动", "组织机构", "网址", "网店", "网络", "谷歌", "集团",
    "飞利浦", "餐厅", "닷넷", "닷컴", "삼성", "onion"]

URL_REGEXP = re.compile((
    r'('
    r'^(?!(https?://|www\.)?\.|ftps?://|([0-9]+\.){{1,3}}\d+)'  # exclude urls that start with "."
    r'(?:https?://|www\.)*^(?!.*@)(?:[\w+-_]+[.])'              # beginning of url
    r'(?:{0}\b'                                                # all tlds
    r'(?:[:0-9]))'                                              # port numbers & close off TLDs
    r'(?:[\w+\/]?[a-z0-9!\*\'\(\);:&=\+\$/%#\[\]\-_\.,~?])*'    # path/query params
    r')').format(r'\b|'.join(TLDS)), re.U | re.I | re.X)


def calc_expected_status_length(status, short_url_length=23):
    """ Calculates the length of a tweet, taking into account Twitter's
    replacement of URLs with https://t.co links.

    Args:
        status: text of the status message to be posted.
        short_url_length: the current published https://t.co links

    Returns:
        Expected length of the status message as an integer.

    """
    status_length = 0
    if isinstance(status, bytes):
        status = unicode(status)
    for word in re.split(r'\s', status):
        if is_url(word):
            status_length += short_url_length
        else:
            for character in word:
                if any([ord(normalize("NFC", character)) in char_range for char_range in CHAR_RANGES]):
                    status_length += 1
                else:
                    status_length += 2
    status_length += len(re.findall(r'\s', status))
    return status_length


def is_url(text):
    """ Checks to see if a bit of text is a URL.

    Args:
        text: text to check.

    Returns:
        Boolean of whether the text should be treated as a URL or not.
    """
    return bool(re.findall(URL_REGEXP, text))


def http_to_file(http):
    data_file = NamedTemporaryFile()
    req = requests.get(http, stream=True)
    for chunk in req.iter_content(chunk_size=1024 * 1024):
        data_file.write(chunk)
    return data_file


def parse_media_file(passed_media, async_upload=False):
    """ Parses a media file and attempts to return a file-like object and
    information about the media file.

    Args:
        passed_media: media file which to parse.
        async_upload: flag, for validation media file attributes.

    Returns:
        file-like object, the filename of the media file, the file size, and
        the type of media.
    """
    img_formats = ['image/jpeg',
                   'image/png',
                   'image/bmp',
                   'image/webp']
    long_img_formats = [
        'image/gif'
    ]
    video_formats = ['video/mp4',
                     'video/quicktime']

    # If passed_media is a string, check if it points to a URL, otherwise,
    # it should point to local file. Create a reference to a file obj for
    #  each case such that data_file ends up with a read() method.
    if not hasattr(passed_media, 'read'):
        if passed_media.startswith('http'):
            data_file = http_to_file(passed_media)
            filename = os.path.basename(urlparse(passed_media).path)
        else:
            data_file = open(os.path.realpath(passed_media), 'rb')
            filename = os.path.basename(passed_media)

    # Otherwise, if a file object was passed in the first place,
    # create the standard reference to media_file (i.e., rename it to fp).
    else:
        if passed_media.mode not in ['rb', 'rb+', 'w+b']:
            raise TwitterError('File mode must be "rb" or "rb+"')
        filename = os.path.basename(passed_media.name)
        data_file = passed_media

    data_file.seek(0, 2)
    file_size = data_file.tell()

    try:
        data_file.seek(0)
    except Exception as e:
        pass

    media_type = mimetypes.guess_type(os.path.basename(filename))[0]
    if media_type is not None:
        if media_type in img_formats and file_size > 5 * 1048576:
            raise TwitterError({'message': 'Images must be less than 5MB.'})
        elif media_type in long_img_formats and file_size > 15 * 1048576:
            raise TwitterError({'message': 'GIF Image must be less than 15MB.'})
        elif media_type in video_formats and not async_upload and file_size > 15 * 1048576:
            raise TwitterError({'message': 'Videos must be less than 15MB.'})
        elif media_type in video_formats and async_upload and file_size > 512 * 1048576:
            raise TwitterError({'message': 'Videos must be less than 512MB.'})
        elif media_type not in img_formats and media_type not in video_formats and media_type not in long_img_formats:
            raise TwitterError({'message': 'Media type could not be determined.'})

    return data_file, filename, file_size, media_type


def enf_type(field, _type, val):
    """ Checks to see if a given val for a field (i.e., the name of the field)
    is of the proper _type. If it is not, raises a TwitterError with a brief
    explanation.

    Args:
        field:
            Name of the field you are checking.
        _type:
            Type that the value should be returned as.
        val:
            Value to convert to _type.

    Returns:
        val converted to type _type.

    """
    try:
        return _type(val)
    except ValueError:
        raise TwitterError({
            'message': '"{0}" must be type {1}'.format(field, _type.__name__)
        })


def parse_arg_list(args, attr):
    out = []
    if isinstance(args, (str, unicode)):
        out.append(args)
    elif isinstance(args, twitter.User):
        out.append(getattr(args, attr))
    elif isinstance(args, (list, tuple)):
        for item in args:
            if isinstance(item, (str, unicode)):
                out.append(item)
            elif isinstance(item, twitter.User):
                out.append(getattr(item, attr))
    return ",".join([str(item) for item in out])



import datetime
import threading

import sickchill.oldbeard.search_queue
from sickchill import logger, settings
from sickchill.helper.exceptions import MultipleShowObjectsException
from sickchill.show.Show import Show

from . import common, db, network_timezones


class DailySearcher(object):  # pylint:disable=too-few-public-methods
    def __init__(self):
        self.lock = threading.Lock()
        self.amActive = False

    def run(self, force=False):  # pylint:disable=too-many-branches
        """
        Runs the daily searcher, queuing selected episodes for search

        :param force: Force search
        """
        if self.amActive:
            return

        self.amActive = True
        logger.info(_("Searching for new released episodes ..."))

        if not network_timezones.network_dict:
            network_timezones.update_network_dict()

        if network_timezones.network_dict:
            curDate = (datetime.date.today() + datetime.timedelta(days=1)).toordinal()
        else:
            curDate = (datetime.date.today() + datetime.timedelta(days=2)).toordinal()

        curTime = datetime.datetime.now(network_timezones.sb_timezone)

        main_db_con = db.DBConnection()
        sql_results = main_db_con.select(
            "SELECT showid, airdate, season, episode FROM tv_episodes WHERE status = ? AND (airdate <= ? and airdate > 1)", [common.UNAIRED, curDate]
        )

        sql_l = []
        show = None

        for sqlEp in sql_results:
            try:
                if not show or int(sqlEp["showid"]) != show.indexerid:
                    show = Show.find(settings.showList, int(sqlEp["showid"]))

                # for when there is orphaned series in the database but not loaded into our showlist
                if not show or show.paused:
                    continue

            except MultipleShowObjectsException:
                logger.info("ERROR: expected to find a single show matching " + str(sqlEp["showid"]))
                continue

            if show.airs and show.network:
                # This is how you assure it is always converted to local time
                air_time = network_timezones.parse_date_time(sqlEp["airdate"], show.airs, show.network).astimezone(network_timezones.sb_timezone)

                # filter out any episodes that haven't started airing yet,
                # but set them to the default status while they are airing
                # so they are snatched faster
                if air_time > curTime:
                    continue

            ep = show.getEpisode(sqlEp["season"], sqlEp["episode"])
            with ep.lock:
                if ep.season == 0:
                    logger.info("New episode " + ep.pretty_name() + " airs today, setting status to SKIPPED because is a special season")
                    ep.status = common.SKIPPED
                else:
                    logger.info(
                        "New episode {0} airs today, setting to default episode status for this show: {1}".format(
                            ep.pretty_name(), common.statusStrings[ep.show.default_ep_status]
                        )
                    )
                    ep.status = ep.show.default_ep_status

                sql_l.append(ep.get_sql())

        if sql_l:
            main_db_con = db.DBConnection()
            main_db_con.mass_action(sql_l)
        else:
            logger.info("No new released episodes found ...")

        # queue episode for daily search
        dailysearch_queue_item = sickchill.oldbeard.search_queue.DailySearchQueueItem()
        settings.searchQueueScheduler.action.add_item(dailysearch_queue_item)

        self.amActive = False



# -*- coding: utf-8 -*-

"""
tmdbsimple.search
~~~~~~~~~~~~~~~~~
This module implements the Search functionality of tmdbsimple.

Created by Celia Oakley on 2013-10-31.

:copyright: (c) 2013-2020 by Celia Oakley
:license: GPLv3, see LICENSE for more details
"""

from .base import TMDB


class Search(TMDB):
    """
    Search functionality

    See: https://developers.themoviedb.org/3/search
    """
    BASE_PATH = 'search'
    URLS = {
        'company': '/company',
        'collection': '/collection',
        'keyword': '/keyword',
        'movie': '/movie',
        'multi': '/multi',
        'person': '/person',
        'tv': '/tv',
    }

    def company(self, **kwargs):
        """
        Search for companies.

        Args:
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('company')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def collection(self, **kwargs):
        """
        Search for collections.

        Args:
            language: (optional) (optional) ISO 639-1 code.
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('collection')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def keyword(self, **kwargs):
        """
        Search for keywords.

        Args:
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('keyword')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def movie(self, **kwargs):
        """
        Search for movies.

        Args:
            language: (optional) (optional) ISO 639-1 code.
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.
            include_adult: (optional) Choose whether to inlcude adult
                (pornography) content in the results.
            region: (optional) Specify a ISO 3166-1 code to filter release
                dates. Must be uppercase.
            year: (optional) A filter to limit the results to a specific year
                (looking at all release dates).
            primary_release_year: (optional) A filter to limit the results to a
                specific primary release year.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('movie')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def multi(self, **kwargs):
        """
        Search multiple models in a single request. Multi search currently
        supports searching for movies, tv shows and people in a single request.

        Args:
            language: (optional) (optional) ISO 639-1 code.
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.
            include_adult: (optional) Choose whether to inlcude adult
                (pornography) content in the results.
            region: (optional) Specify a ISO 3166-1 code to filter release
                dates. Must be uppercase.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('multi')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def person(self, **kwargs):
        """
        Search for people.

        Args:
            language: (optional) (optional) ISO 639-1 code.
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.
            include_adult: (optional) Choose whether to inlcude adult
                (pornography) content in the results.
            region: (optional) Specify a ISO 3166-1 code to filter release
                dates. Must be uppercase.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('person')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response

    def tv(self, **kwargs):
        """
        Search for a TV show.

        Args:
            language: (optional) (optional) ISO 639-1 code.
            query: (required) Pass a text query to search. This value should be
                URI encoded.
            page: (optional) Minimum 1, maximum 1000, default 1.
            include_adult: (optional) Choose whether to inlcude adult
                (pornography) content in the results.
            first_air_date_year: (optional) Filter the results to only match
                shows that have an air date with with value.

        Returns:
            A dict respresentation of the JSON returned from the API.
        """
        path = self._get_path('tv')

        response = self._GET(path, kwargs)
        self._set_attrs_to_values(response)
        return response



# https://github.com/urbanguacamole/torrent-paradise

from sickchill import logger
from sickchill.helper.common import try_int
from sickchill.oldbeard import tvcache
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        super().__init__("torrent-paradise")
        self.minseed = 0
        self.minleech = 0
        self.supports_movies = True
        self.ability_status = self.PROVIDER_BACKLOG

        self.url = "https://torrent-paradise.ml"
        self.urls = {"search": "https://torrent-paradise.ml/api/search"}

        self.cache = tvcache.TVCache(self)

    def search(self, search_strings, age=0, ep_obj=None):
        results = []

        if "RSS" in search_strings:
            del search_strings["RSS"]

        for mode in search_strings:
            items = []

            search_string = " OR ".join(search_strings[mode])
            logger.debug(_("Search String: {search_string}".format(search_string=search_string)))

            data = self.get_url(self.urls["search"], params={"q": search_string}, returns="json")

            if not isinstance(data, list):
                logger.debug("No data returned from provider")
                continue

            if not data:
                logger.debug("Data returned from provider does not contain any torrents")
                continue

            logger.info(f"Number of torrents found on {self.name} = {len(data)}")

            for item in data:
                try:
                    title = item.pop("text")
                    info_hash = item.pop("id")
                    if not all([title, info_hash]):
                        continue

                    seeders = item.pop("s")
                    leechers = item.pop("l")

                    if seeders < self.minseed or leechers < self.minleech:
                        logger.debug(f"Discarding torrent because it doesn't meet the minimum seeders or leechers: {title} (S:{seeders} L:{leechers})")
                        continue

                    size = item.pop("len")
                    logger.debug(f"Found result: {title} with {seeders} seeders and {leechers} leechers with a file size {size}")

                    items.append(
                        {
                            "title": title,
                            "size": size,
                            "seeders": seeders,
                            "leechers": leechers,
                            "hash": info_hash,
                            "link": f"magnet:?xt=urn:btih:{info_hash}&dn={title}{self._custom_trackers}",
                        }
                    )
                except Exception as error:
                    raise error

            # For each search mode sort all the items by seeders
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items
        return results



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2018-11-11

@author: marcel-odya
'''

base = [
    ["właśnie teraz", "za chwilę"],
    ["%s sekund temu", "za %s sekund", "%s sekundy temu", "za %s sekundy"],
    ["minutę temu", "za minutę"],
    ["%s minut temu", "za %s minut", "%s minuty temu", "za %s minuty"],
    ["godzinę temu", "za godzinę"],
    ["%s godzin temu", "za %s godzin", "%s godziny temu", "za %s godziny"],
    ["1 dzień temu", "za 1 dzień"],
    ["%s dni temu", "za %s dni", "%s dni temu", "za %s dni"],
    ["tydzień temu", "za tydzień"],
    ["%s tygodni temu", "za %s tygodni", "%s tygodnie temu", "za %s tygodnie"],
    ["miesiąc temu", "za miesiąc"],
    ["%s miesięcy temu", "za %s miesięcy", "%s miesiące temu", "za %s miesiące"],
    ["rok temu", "za rok"],
    ["%s lat temu", "za %s lat", "%s lata temu", "za %s lata"]
]


def generate(row, y):
    def formatting(time):
        '''
        Uses the 3rd and 4th field of the list in every 2 entries -
        the ones containing %s, if the diff ends with 2, 3 or 4 but
        not with 12, 13 or 14.
        '''
        if row % 2 == 0:
            return base[row][y]
        last_number = time % 10
        last_two_numbers = time % 100
        if last_number in range(2, 5) and last_two_numbers not in range(12, 15):
            return base[row][y + 2]
        return base[row][y]

    return formatting


LOCALE = generate



import json
from fanart.errors import ResponseFanartError
import os
import unittest
from httpretty import HTTPretty, httprettified
from fanart.tv import *
from fanart.tests import LOCALDIR
os.environ['FANART_APIKEY'] = 'e3c7f0d0beeaf45b3a0dd3b9dd8a3338'


class TvItemTestCase(unittest.TestCase):
    @httprettified
    def test_get_wilfred(self):
        with open(os.path.join(LOCALDIR, 'response/tv_239761.json')) as fp:
            body = fp.read()
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://webservice.fanart.tv/v3/tv/239761?api_key={}'.format(
                os.environ['FANART_APIKEY']),
            body=body
        )
        wilfred = TvShow.get(id=239761)

        # If we update `response/tv_239761.json`, then we also must update
        # `json/wilfred.json`, and to do so, we can use the following command:
        # print(wilfred.json(indent=4))

        self.assertEqual(wilfred.tvdbid, '239761')
        with open(os.path.join(LOCALDIR, 'json/wilfred.json')) as fp:
            self.assertEqual(json.loads(wilfred.json()), json.load(fp))

    @httprettified
    def test_get_dexter(self):
        with open(os.path.join(LOCALDIR, 'response/tv_79349.json')) as fp:
            body = fp.read()
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://webservice.fanart.tv/v3/tv/79349?api_key={}'.format(
                os.environ['FANART_APIKEY']),
            body=body
        )
        dexter = TvShow.get(id=79349)
        self.assertEqual(dexter.tvdbid, '79349')
        self.assertEqual(dexter, eval(repr(dexter)))

    @httprettified
    def test_get_null(self):
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://webservice.fanart.tv/v3/tv/79349?api_key={}'.format(
                os.environ['FANART_APIKEY']),
            body='null'
        )
        self.assertRaises(ResponseFanartError, TvShow.get, id=79349)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2017-8-30

@author: generated by @lolobosse script
'''
LOCALE = [
    ["à l'instant", "dans un instant"],
    ["il y a %s secondes", "dans %s secondes"],
    ["il y a 1 minute", "dans 1 minute"],
    ["il y a %s minutes", "dans %s minutes"],
    ["il y a 1 heure", "dans 1 heure"],
    ["il y a %s heures", "dans %s heures"],
    ["il y a 1 jour", "dans 1 jour"],
    ["il y a %s jours", "dans %s jours"],
    ["il y a 1 semaine", "dans 1 semaine"],
    ["il y a %s semaines", "dans %s semaines"],
    ["il y a 1 mois", "dans 1 mois"],
    ["il y a %s mois", "dans %s mois"],
    ["il y a 1 an", "dans 1 an"],
    ["il y a %s ans", "dans %s ans"]
]
        


import ast
import smtplib
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.utils import formatdate

from sickchill import logger, settings
from sickchill.oldbeard import db


class Notifier(object):
    def __init__(self):
        self.last_err = None

    def test_notify(self, host, port, smtp_from, use_tls, user, pwd, to):
        msg = MIMEText("This is a test message from SickChill.  If you're reading this, the test succeeded.")
        if settings.EMAIL_SUBJECT:
            msg["Subject"] = "[TEST] " + settings.EMAIL_SUBJECT
        else:
            msg["Subject"] = "SickChill: Test Message"

        msg["From"] = smtp_from
        msg["To"] = to
        msg["Date"] = formatdate(localtime=True)
        return self._sendmail(host, port, smtp_from, use_tls, user, pwd, [to], msg, True)

    def notify_snatch(self, ep_name, title="Snatched:"):
        """
        Send a notification that an episode was snatched

        ep_name: The name of the episode that was snatched
        title: The title of the notification (optional)
        """
        if settings.USE_EMAIL and settings.EMAIL_NOTIFY_ONSNATCH:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if not to:
                logger.debug("Skipping email notify because there are no configured recipients")
            else:
                try:
                    msg = MIMEMultipart("alternative")
                    msg.attach(
                        MIMEText(
                            "SickChill Notification - Snatched\n"
                            "Show: {0}\nEpisode Number: {1}\nEpisode: {2}\nQuality: {3}\n\n"
                            "Powered by SickChill.".format(show[0], show[1], show[2], show[3])
                        )
                    )
                    msg.attach(
                        MIMEText(
                            '<body style="font-family:Helvetica, Arial, sans-serif;">'
                            "<h3>SickChill Notification - Snatched</h3>"
                            "<p>Show: <b>{0}</b></p><p>Episode Number: <b>{1}</b></p><p>Episode: <b>{2}</b></p><p>Quality: <b>{3}</b></p>"
                            '<h5 style="margin-top: 2.5em; padding: .7em 0; '
                            'color: #777; border-top: #BBB solid 1px;">'
                            "Powered by SickChill.</h5></body>".format(show[0], show[1], show[2], show[3]),
                            "html",
                        )
                    )

                except Exception:
                    try:
                        msg = MIMEText(ep_name)
                    except Exception:
                        msg = MIMEText("Episode Snatched")

                if settings.EMAIL_SUBJECT:
                    msg["Subject"] = "[SN] " + settings.EMAIL_SUBJECT
                else:
                    msg["Subject"] = "Snatched: " + ep_name
                msg["From"] = settings.EMAIL_FROM
                msg["To"] = ",".join(to)
                msg["Date"] = formatdate(localtime=True)
                if self._sendmail(
                    settings.EMAIL_HOST, settings.EMAIL_PORT, settings.EMAIL_FROM, settings.EMAIL_TLS, settings.EMAIL_USER, settings.EMAIL_PASSWORD, to, msg
                ):
                    logger.debug('Snatch notification sent to [{0}] for "{1}"'.format(to, ep_name))
                else:
                    logger.warning("Snatch notification error: {0}".format(self.last_err))

    def notify_download(self, ep_name, title="Completed:"):
        """
        Send a notification that an episode was downloaded

        ep_name: The name of the episode that was downloaded
        title: The title of the notification (optional)
        """
        if settings.USE_EMAIL and settings.EMAIL_NOTIFY_ONDOWNLOAD:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if not to:
                logger.debug("Skipping email notify because there are no configured recipients")
            else:
                try:
                    msg = MIMEMultipart("alternative")
                    msg.attach(
                        MIMEText(
                            "SickChill Notification - Downloaded\n"
                            "Show: {0}\nEpisode Number: {1}\nEpisode: {2}\nQuality: {3}\n\n"
                            "Powered by SickChill.".format(show[0], show[1], show[2], show[3])
                        )
                    )
                    msg.attach(
                        MIMEText(
                            '<body style="font-family:Helvetica, Arial, sans-serif;">'
                            "<h3>SickChill Notification - Downloaded</h3>"
                            "<p>Show: <b>{0}</b></p><p>Episode Number: <b>{1}</b></p><p>Episode: <b>{2}</b></p><p>Quality: <b>{3}</b></p>"
                            '<h5 style="margin-top: 2.5em; padding: .7em 0; '
                            'color: #777; border-top: #BBB solid 1px;">'
                            "Powered by SickChill.</h5></body>".format(show[0], show[1], show[2], show[3]),
                            "html",
                        )
                    )

                except Exception:
                    try:
                        msg = MIMEText(ep_name)
                    except Exception:
                        msg = MIMEText("Episode Downloaded")

                if settings.EMAIL_SUBJECT:
                    msg["Subject"] = "[DL] " + settings.EMAIL_SUBJECT
                else:
                    msg["Subject"] = "Downloaded: " + ep_name
                msg["From"] = settings.EMAIL_FROM
                msg["To"] = ",".join(to)
                msg["Date"] = formatdate(localtime=True)
                if self._sendmail(
                    settings.EMAIL_HOST, settings.EMAIL_PORT, settings.EMAIL_FROM, settings.EMAIL_TLS, settings.EMAIL_USER, settings.EMAIL_PASSWORD, to, msg
                ):
                    logger.debug('Download notification sent to [{0}] for "{1}"'.format(to, ep_name))
                else:
                    logger.warning("Download notification error: {0}".format(self.last_err))

    def notify_postprocess(self, ep_name, title="Postprocessed:"):
        """
        Send a notification that an episode was postprocessed

        ep_name: The name of the episode that was postprocessed
        title: The title of the notification (optional)
        """
        if settings.USE_EMAIL and settings.EMAIL_NOTIFY_ONPOSTPROCESS:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if not to:
                logger.debug("Skipping email notify because there are no configured recipients")
            else:
                try:
                    msg = MIMEMultipart("alternative")
                    msg.attach(
                        MIMEText(
                            "SickChill Notification - Postprocessed\n"
                            "Show: {0}\nEpisode Number: {1}\nEpisode: {2}\nQuality: {3}\n\n"
                            "Powered by SickChill.".format(show[0], show[1], show[2], show[3])
                        )
                    )
                    msg.attach(
                        MIMEText(
                            '<body style="font-family:Helvetica, Arial, sans-serif;">'
                            "<h3>SickChill Notification - Postprocessed</h3>"
                            "<p>Show: <b>{0}</b></p><p>Episode Number: <b>{1}</b></p><p>Episode: <b>{2}</b></p><p>Quality: <b>{3}</b></p>"
                            '<h5 style="margin-top: 2.5em; padding: .7em 0; '
                            'color: #777; border-top: #BBB solid 1px;">'
                            "Powered by SickChill.</h5></body>".format(show[0], show[1], show[2], show[3]),
                            "html",
                        )
                    )

                except Exception:
                    try:
                        msg = MIMEText(ep_name)
                    except Exception:
                        msg = MIMEText("Episode Postprocessed")

                if settings.EMAIL_SUBJECT:
                    msg["Subject"] = "[PP] " + settings.EMAIL_SUBJECT
                else:
                    msg["Subject"] = "Postprocessed: " + ep_name
                msg["From"] = settings.EMAIL_FROM
                msg["To"] = ",".join(to)
                msg["Date"] = formatdate(localtime=True)
                if self._sendmail(
                    settings.EMAIL_HOST, settings.EMAIL_PORT, settings.EMAIL_FROM, settings.EMAIL_TLS, settings.EMAIL_USER, settings.EMAIL_PASSWORD, to, msg
                ):
                    logger.debug('Postprocess notification sent to [{0}] for "{1}"'.format(to, ep_name))
                else:
                    logger.warning("Postprocess notification error: {0}".format(self.last_err))

    def notify_subtitle_download(self, ep_name, lang, title="Downloaded subtitle:"):
        """
        Send a notification that an subtitle was downloaded

        ep_name: The name of the episode that was downloaded
        lang: Subtitle language wanted
        """
        if settings.USE_EMAIL and settings.EMAIL_NOTIFY_ONSUBTITLEDOWNLOAD:
            show = self._parseEp(ep_name)
            to = self._generate_recipients(show)
            if not to:
                logger.debug("Skipping email notify because there are no configured recipients")
            else:
                try:
                    msg = MIMEMultipart("alternative")
                    msg.attach(
                        MIMEText(
                            "SickChill Notification - Subtitle Downloaded\n"
                            "Show: {0}\nEpisode Number: {1}\nEpisode: {2}\n"
                            "Language: {3}\n\n"
                            "Powered by SickChill.".format(show[0], show[1], show[2], lang)
                        )
                    )
                    msg.attach(
                        MIMEText(
                            '<body style="font-family:Helvetica, Arial, sans-serif;">'
                            "<h3>SickChill Notification - Subtitle Downloaded</h3>"
                            "<p>Show: <b>{0}</b></p><p>Episode Number: <b>{1}</b></p><p>Episode: <b>{2}</b></p></p>"
                            "<p>Language: <b>{3}</b></p>"
                            '<h5 style="margin-top: 2.5em; padding: .7em 0; '
                            'color: #777; border-top: #BBB solid 1px;">'
                            "Powered by SickChill.</h5></body>".format(show[0], show[1], show[2], lang),
                            "html",
                        )
                    )
                except Exception:
                    try:
                        msg = MIMEText(ep_name + ": " + lang)
                    except Exception:
                        msg = MIMEText("Episode Subtitle Downloaded")

                if settings.EMAIL_SUBJECT:
                    msg["Subject"] = "[ST] " + settings.EMAIL_SUBJECT
                else:
                    msg["Subject"] = lang + " Subtitle Downloaded: " + ep_name
                msg["From"] = settings.EMAIL_FROM
                msg["To"] = ",".join(to)
                if self._sendmail(
                    settings.EMAIL_HOST, settings.EMAIL_PORT, settings.EMAIL_FROM, settings.EMAIL_TLS, settings.EMAIL_USER, settings.EMAIL_PASSWORD, to, msg
                ):
                    logger.debug('Download notification sent to [{0}] for "{1}"'.format(to, ep_name))
                else:
                    logger.warning("Download notification error: {0}".format(self.last_err))

    def notify_git_update(self, new_version="??"):
        """
        Send a notification that SickChill was updated
        new_version: The commit SickChill was updated to
        """
        if settings.USE_EMAIL:
            to = self._generate_recipients(None)
            if not to:
                logger.debug("Skipping email notify because there are no configured recipients")
            else:
                try:
                    msg = MIMEMultipart("alternative")
                    msg.attach(MIMEText("SickChill Notification - Updated\n" "Commit: {}\n\n" "Powered by SickChill.".format(new_version)))
                    msg.attach(
                        MIMEText(
                            '<body style="font-family:Helvetica, Arial, sans-serif;">'
                            "<h3>SickChill Notification - Updated</h3><br>"
                            "<p>Commit: <b>{}</b></p><br><br>"
                            '<footer style="margin-top: 2.5em; padding: .7em 0; '
                            'color: #777; border-top: #BBB solid 1px;">'
                            "Powered by SickChill.</footer></body>".format(new_version),
                            "html",
                        )
                    )

                except Exception:
                    try:
                        msg = MIMEText(new_version)
                    except Exception:
                        msg = MIMEText("SickChill updated")

                msg["Subject"] = "Updated: {0}".format(new_version)
                msg["From"] = settings.EMAIL_FROM
                msg["To"] = ",".join(to)
                msg["Date"] = formatdate(localtime=True)
                if self._sendmail(
                    settings.EMAIL_HOST, settings.EMAIL_PORT, settings.EMAIL_FROM, settings.EMAIL_TLS, settings.EMAIL_USER, settings.EMAIL_PASSWORD, to, msg
                ):
                    logger.debug("Update notification sent to [{0}]".format(to))
                else:
                    logger.warning("Update notification error: {0}".format(self.last_err))

    def notify_login(self, ipaddress=""):
        """
        Send a notification that SickChill was logged into remotely
        ipaddress: The ip SickChill was logged into from
        """
        if settings.USE_EMAIL:
            to = self._generate_recipients(None)
            if not len(to):
                logger.debug("Skipping email notify because there are no configured recipients")
            else:
                try:
                    msg = MIMEMultipart("alternative")
                    msg.attach(MIMEText("SickChill Notification - Remote Login\n" "New login from IP: {0}\n\n" "Powered by SickChill.".format(ipaddress)))
                    msg.attach(
                        MIMEText(
                            '<body style="font-family:Helvetica, Arial, sans-serif;">'
                            "<h3>SickChill Notification - Remote Login</h3><br>"
                            '<p>New login from IP: <a href="http://geomaplookup.net/?ip={0}">{0}</a>.<br><br>'
                            '<footer style="margin-top: 2.5em; padding: .7em 0; '
                            'color: #777; border-top: #BBB solid 1px;">'
                            "Powered by SickChill.</footer></body>".format(ipaddress),
                            "html",
                        )
                    )

                except Exception:
                    try:
                        msg = MIMEText(ipaddress)
                    except Exception:
                        msg = MIMEText("SickChill Remote Login")

                msg["Subject"] = "New Login from IP: {0}".format(ipaddress)
                msg["From"] = settings.EMAIL_FROM
                msg["To"] = ",".join(to)
                msg["Date"] = formatdate(localtime=True)
                if self._sendmail(
                    settings.EMAIL_HOST, settings.EMAIL_PORT, settings.EMAIL_FROM, settings.EMAIL_TLS, settings.EMAIL_USER, settings.EMAIL_PASSWORD, to, msg
                ):
                    logger.debug("Login notification sent to [{0}]".format(to))
                else:
                    logger.warning("Login notification error: {0}".format(self.last_err))

    @staticmethod
    def _generate_recipients(show):
        addrs = []
        main_db_con = db.DBConnection()

        # Grab the global recipients
        if settings.EMAIL_LIST:
            for addr in settings.EMAIL_LIST.split(","):
                if len(addr.strip()) > 0:
                    addrs.append(addr)

        # Grab the per-show-notification recipients
        if show is not None:
            for s in show:
                for subs in main_db_con.select("SELECT notify_list FROM tv_shows WHERE show_name = ?", (s,)):
                    if subs["notify_list"]:
                        if subs["notify_list"][0] == "{":
                            entries = dict(ast.literal_eval(subs["notify_list"]))
                            for addr in entries["emails"].split(","):
                                if len(addr.strip()) > 0:
                                    addrs.append(addr)
                        else:  # Legacy
                            for addr in subs["notify_list"].split(","):
                                if len(addr.strip()) > 0:
                                    addrs.append(addr)

        addrs = set(addrs)
        logger.debug("Notification recipients: {0}".format(addrs))
        return addrs

    def _sendmail(self, host, port, smtp_from, use_tls, user, pwd, to, msg, smtpDebug=False):
        logger.debug("HOST: {0}; PORT: {1}; FROM: {2}, TLS: {3}, USER: {4}, PWD: {5}, TO: {6}".format(host, port, smtp_from, use_tls, user, pwd, to))
        try:
            srv = smtplib.SMTP(host, int(port))
        except Exception as e:
            logger.warning("Exception generated while sending e-mail: " + str(e))
            # logger.debug(traceback.format_exc())
            self.last_err = "{0}".format(e)
            return False

        if smtpDebug:
            srv.set_debuglevel(1)
        try:
            if use_tls in ("1", True) or (user and pwd):
                logger.debug("Sending initial EHLO command!")
                srv.ehlo()
            if use_tls in ("1", True):
                logger.debug("Sending STARTTLS command!")
                srv.starttls()
                srv.ehlo()
            if user and pwd:
                logger.debug("Sending LOGIN command!")
                srv.login(user, pwd)

            srv.sendmail(smtp_from, to, msg.as_string())
            srv.quit()
            return True
        except Exception as e:
            self.last_err = "{0}".format(e)
            return False

    @staticmethod
    def _parseEp(ep_name):
        sep = " - "
        titles = ep_name.split(sep)
        logger.debug("TITLES: {0}".format(titles))
        return titles



# Author: Marvin Pinto <me@marvinp.ca>
# Author: Dennis Lutter <lad1337@gmail.com>
import urllib.parse
import urllib.request

from sickchill import logger, settings
from sickchill.oldbeard.common import (
    NOTIFY_DOWNLOAD,
    NOTIFY_GIT_UPDATE,
    NOTIFY_GIT_UPDATE_TEXT,
    NOTIFY_LOGIN,
    NOTIFY_LOGIN_TEXT,
    NOTIFY_SNATCH,
    NOTIFY_SUBTITLE_DOWNLOAD,
    notifyStrings,
)


class Notifier(object):
    def test_notify(self, cust_id=None, apiKey=None):
        return self._notifyFreeMobile("Test", "This is a test notification from SickChill", cust_id, apiKey, force=True)

    @staticmethod
    def _sendFreeMobileSMS(title, msg, cust_id=None, apiKey=None):
        """
        Sends a SMS notification

        msg: The message to send (str)
        title: The title of the message
        userKey: The pushover user id to send the message to (or to subscribe with)

        returns: True if the message succeeded, False otherwise
        """

        if cust_id is None:
            cust_id = settings.FREEMOBILE_ID
        if apiKey is None:
            apiKey = settings.FREEMOBILE_APIKEY

        logger.debug("Free Mobile in use with API KEY: " + apiKey)

        # build up the URL and parameters
        msg = msg.strip()
        msg_quoted = urllib.parse.quote(title + ": " + msg)
        URL = "https://smsapi.free-mobile.fr/sendmsg?user=" + cust_id + "&pass=" + apiKey + "&msg=" + msg_quoted

        req = urllib.request.Request(URL)
        # send the request to Free Mobile
        try:
            urllib.request.urlopen(req)
        except IOError as e:
            if hasattr(e, "code"):
                if e.code == 400:
                    message = "Missing parameter(s)."
                    logger.exception(message)
                    return False, message
                if e.code == 402:
                    message = "Too much SMS sent in a short time."
                    logger.exception(message)
                    return False, message
                if e.code == 403:
                    message = "API service isn't enabled in your account or ID / API key is incorrect."
                    logger.exception(message)
                    return False, message
                if e.code == 500:
                    message = "Server error. Please retry in few moment."
                    logger.exception(message)
                    return False, message
        except Exception as e:
            message = "Error while sending SMS: {0}".format(e)
            logger.exception(message)
            return False, message

        message = "Free Mobile SMS successful."
        logger.info(message)
        return True, message

    def notify_snatch(self, ep_name, title=notifyStrings[NOTIFY_SNATCH]):
        if settings.FREEMOBILE_NOTIFY_ONSNATCH:
            self._notifyFreeMobile(title, ep_name)

    def notify_download(self, ep_name, title=notifyStrings[NOTIFY_DOWNLOAD]):
        if settings.FREEMOBILE_NOTIFY_ONDOWNLOAD:
            self._notifyFreeMobile(title, ep_name)

    def notify_subtitle_download(self, ep_name, lang, title=notifyStrings[NOTIFY_SUBTITLE_DOWNLOAD]):
        if settings.FREEMOBILE_NOTIFY_ONSUBTITLEDOWNLOAD:
            self._notifyFreeMobile(title, ep_name + ": " + lang)

    def notify_git_update(self, new_version="??"):
        if settings.USE_FREEMOBILE:
            update_text = notifyStrings[NOTIFY_GIT_UPDATE_TEXT]
            title = notifyStrings[NOTIFY_GIT_UPDATE]
            self._notifyFreeMobile(title, update_text + new_version)

    def notify_login(self, ipaddress=""):
        if settings.USE_FREEMOBILE:
            update_text = notifyStrings[NOTIFY_LOGIN_TEXT]
            title = notifyStrings[NOTIFY_LOGIN]
            self._notifyFreeMobile(title, update_text.format(ipaddress))

    def _notifyFreeMobile(self, title, message, cust_id=None, apiKey=None, force=False):
        """
        Sends a SMS notification

        title: The title of the notification to send
        message: The message string to send
        cust_id: Your Free Mobile customer ID
        apikey: Your Free Mobile API key
        force: Enforce sending, for instance for testing
        """

        if not settings.USE_FREEMOBILE and not force:
            logger.debug("Notification for Free Mobile not enabled, skipping this notification")
            return False, "Disabled"

        logger.debug("Sending a SMS for " + message)

        return self._sendFreeMobileSMS(title, message, cust_id, apiKey)



# -*- coding: utf-8 -*-
from pkg_resources import EntryPoint

from stevedore import ExtensionManager


class RegistrableExtensionManager(ExtensionManager):
    """:class:~stevedore.extensions.ExtensionManager` with support for registration.

    It allows loading of internal extensions without setup and registering/unregistering additional extensions.

    Loading is done in this order:

    * Entry point extensions
    * Internal extensions
    * Registered extensions

    :param str namespace: namespace argument for :class:~stevedore.extensions.ExtensionManager`.
    :param list internal_extensions: internal extensions to use with entry point syntax.
    :param \*\*kwargs: additional parameters for the :class:~stevedore.extensions.ExtensionManager` constructor.

    """
    def __init__(self, namespace, internal_extensions, **kwargs):
        #: Registered extensions with entry point syntax
        self.registered_extensions = []

        #: Internal extensions with entry point syntax
        self.internal_extensions = internal_extensions

        super(RegistrableExtensionManager, self).__init__(namespace, **kwargs)

    def list_entry_points(self):
        # copy of default extensions
        eps = list(super(RegistrableExtensionManager, self).list_entry_points())

        # internal extensions
        for iep in self.internal_extensions:
            ep = EntryPoint.parse(iep)
            if ep.name not in [e.name for e in eps]:
                eps.append(ep)

        # registered extensions
        for rep in self.registered_extensions:
            ep = EntryPoint.parse(rep)
            if ep.name not in [e.name for e in eps]:
                eps.append(ep)

        return eps

    def register(self, entry_point):
        """Register an extension

        :param str entry_point: extension to register (entry point syntax).
        :raise: ValueError if already registered.

        """
        if entry_point in self.registered_extensions:
            raise ValueError('Extension already registered')

        ep = EntryPoint.parse(entry_point)
        if ep.name in self.names():
            raise ValueError('An extension with the same name already exist')

        ext = self._load_one_plugin(ep, False, (), {}, False)
        self.extensions.append(ext)
        if self._extensions_by_name is not None:
            self._extensions_by_name[ext.name] = ext
        self.registered_extensions.insert(0, entry_point)

    def unregister(self, entry_point):
        """Unregister a provider

        :param str entry_point: provider to unregister (entry point syntax).

        """
        if entry_point not in self.registered_extensions:
            raise ValueError('Extension not registered')

        ep = EntryPoint.parse(entry_point)
        self.registered_extensions.remove(entry_point)
        if self._extensions_by_name is not None:
            del self._extensions_by_name[ep.name]
        for i, ext in enumerate(self.extensions):
            if ext.name == ep.name:
                del self.extensions[i]
                break


#: Provider manager
provider_manager = RegistrableExtensionManager('subliminal.providers', [
    'addic7ed = subliminal.providers.addic7ed:Addic7edProvider',
    'argenteam = subliminal.providers.argenteam:ArgenteamProvider',
    'legendastv = subliminal.providers.legendastv:LegendasTVProvider',
    'opensubtitles = subliminal.providers.opensubtitles:OpenSubtitlesProvider',
    'opensubtitlesvip = subliminal.providers.opensubtitles:OpenSubtitlesVipProvider',
    'podnapisi = subliminal.providers.podnapisi:PodnapisiProvider',
    'shooter = subliminal.providers.shooter:ShooterProvider',
    'thesubdb = subliminal.providers.thesubdb:TheSubDBProvider',
    'tvsubtitles = subliminal.providers.tvsubtitles:TVsubtitlesProvider'
])

#: Disabled providers
disabled_providers = ['opensubtitlesvip']

#: Default enabled providers
default_providers = [p for p in provider_manager.names() if p not in disabled_providers]

#: Refiner manager
refiner_manager = RegistrableExtensionManager('subliminal.refiners', [
    'hash = subliminal.refiners.hash:refine',
    'metadata = subliminal.refiners.metadata:refine',
    'omdb = subliminal.refiners.omdb:refine',
    'tvdb = subliminal.refiners.tvdb:refine'
])



# ------------------------------------------------------------------------------- #

import logging
import re
import requests
import sys
import ssl

from collections import OrderedDict
from copy import deepcopy

from requests.adapters import HTTPAdapter
from requests.sessions import Session
from requests_toolbelt.utils import dump

from time import sleep

# ------------------------------------------------------------------------------- #

try:
    import brotli
except ImportError:
    pass

try:
    import copyreg
except ImportError:
    import copy_reg as copyreg

try:
    from HTMLParser import HTMLParser
except ImportError:
    if sys.version_info >= (3, 4):
        import html
    else:
        from html.parser import HTMLParser

try:
    from urlparse import urlparse, urljoin
except ImportError:
    from urllib.parse import urlparse, urljoin

# ------------------------------------------------------------------------------- #

from .exceptions import (
    CloudflareLoopProtection,
    CloudflareCode1020,
    CloudflareIUAMError,
    CloudflareSolveError,
    CloudflareChallengeError,
    CloudflareCaptchaError,
    CloudflareCaptchaProvider
)

from .interpreters import JavaScriptInterpreter
from .captcha import Captcha
from .user_agent import User_Agent

# ------------------------------------------------------------------------------- #

__version__ = '1.2.56'

# ------------------------------------------------------------------------------- #


class CipherSuiteAdapter(HTTPAdapter):

    __attrs__ = [
        'ssl_context',
        'max_retries',
        'config',
        '_pool_connections',
        '_pool_maxsize',
        '_pool_block',
        'source_address'
    ]

    def __init__(self, *args, **kwargs):
        self.ssl_context = kwargs.pop('ssl_context', None)
        self.cipherSuite = kwargs.pop('cipherSuite', None)
        self.source_address = kwargs.pop('source_address', None)

        if self.source_address:
            if isinstance(self.source_address, str):
                self.source_address = (self.source_address, 0)

            if not isinstance(self.source_address, tuple):
                raise TypeError(
                    "source_address must be IP address string or (ip, port) tuple"
                )

        if not self.ssl_context:
            self.ssl_context = ssl.create_default_context(ssl.Purpose.SERVER_AUTH)
            self.ssl_context.set_ciphers(self.cipherSuite)
            self.ssl_context.set_ecdh_curve('prime256v1')
            self.ssl_context.options |= (ssl.OP_NO_SSLv2 | ssl.OP_NO_SSLv3 | ssl.OP_NO_TLSv1 | ssl.OP_NO_TLSv1_1)

        super(CipherSuiteAdapter, self).__init__(**kwargs)

    # ------------------------------------------------------------------------------- #

    def init_poolmanager(self, *args, **kwargs):
        kwargs['ssl_context'] = self.ssl_context
        kwargs['source_address'] = self.source_address
        return super(CipherSuiteAdapter, self).init_poolmanager(*args, **kwargs)

    # ------------------------------------------------------------------------------- #

    def proxy_manager_for(self, *args, **kwargs):
        kwargs['ssl_context'] = self.ssl_context
        kwargs['source_address'] = self.source_address
        return super(CipherSuiteAdapter, self).proxy_manager_for(*args, **kwargs)

# ------------------------------------------------------------------------------- #


class CloudScraper(Session):

    def __init__(self, *args, **kwargs):
        self.debug = kwargs.pop('debug', False)
        self.delay = kwargs.pop('delay', None)
        self.cipherSuite = kwargs.pop('cipherSuite', None)
        self.ssl_context = kwargs.pop('ssl_context', None)
        self.interpreter = kwargs.pop('interpreter', 'native')
        self.captcha = kwargs.pop('captcha', {})
        self.requestPreHook = kwargs.pop('requestPreHook', None)
        self.requestPostHook = kwargs.pop('requestPostHook', None)
        self.source_address = kwargs.pop('source_address', None)
        self.doubleDown = kwargs.pop('doubleDown', True)

        self.allow_brotli = kwargs.pop(
            'allow_brotli',
            True if 'brotli' in sys.modules.keys() else False
        )

        self.user_agent = User_Agent(
            allow_brotli=self.allow_brotli,
            browser=kwargs.pop('browser', None)
        )

        self._solveDepthCnt = 0
        self.solveDepth = kwargs.pop('solveDepth', 3)

        super(CloudScraper, self).__init__(*args, **kwargs)

        # pylint: disable=E0203
        if 'requests' in self.headers['User-Agent']:
            # ------------------------------------------------------------------------------- #
            # Set a random User-Agent if no custom User-Agent has been set
            # ------------------------------------------------------------------------------- #
            self.headers = self.user_agent.headers
            if not self.cipherSuite:
                self.cipherSuite = self.user_agent.cipherSuite

        if isinstance(self.cipherSuite, list):
            self.cipherSuite = ':'.join(self.cipherSuite)

        self.mount(
            'https://',
            CipherSuiteAdapter(
                cipherSuite=self.cipherSuite,
                ssl_context=self.ssl_context,
                source_address=self.source_address
            )
        )

        # purely to allow us to pickle dump
        copyreg.pickle(ssl.SSLContext, lambda obj: (obj.__class__, (obj.protocol,)))

    # ------------------------------------------------------------------------------- #
    # Allow us to pickle our session back with all variables
    # ------------------------------------------------------------------------------- #

    def __getstate__(self):
        return self.__dict__

    # ------------------------------------------------------------------------------- #
    # Allow replacing actual web request call via subclassing
    # ------------------------------------------------------------------------------- #

    def perform_request(self, method, url, *args, **kwargs):
        return super(CloudScraper, self).request(method, url, *args, **kwargs)

    # ------------------------------------------------------------------------------- #
    # Raise an Exception with no stacktrace and reset depth counter.
    # ------------------------------------------------------------------------------- #

    def simpleException(self, exception, msg):
        self._solveDepthCnt = 0
        sys.tracebacklimit = 0
        raise exception(msg)

    # ------------------------------------------------------------------------------- #
    # debug the request via the response
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def debugRequest(req):
        try:
            print(dump.dump_all(req).decode('utf-8'))
        except ValueError as e:
            print(f"Debug Error: {getattr(e, 'message', e)}")

    # ------------------------------------------------------------------------------- #
    # Unescape / decode html entities
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def unescape(html_text):
        if sys.version_info >= (3, 0):
            if sys.version_info >= (3, 4):
                return html.unescape(html_text)

            return HTMLParser().unescape(html_text)

        return HTMLParser().unescape(html_text)

    # ------------------------------------------------------------------------------- #
    # Decode Brotli on older versions of urllib3 manually
    # ------------------------------------------------------------------------------- #

    def decodeBrotli(self, resp):
        if requests.packages.urllib3.__version__ < '1.25.1' and resp.headers.get('Content-Encoding') == 'br':
            if self.allow_brotli and resp._content:
                resp._content = brotli.decompress(resp.content)
            else:
                logging.warning(
                    f'You\'re running urllib3 {requests.packages.urllib3.__version__}, Brotli content detected, '
                    'Which requires manual decompression, '
                    'But option allow_brotli is set to False, '
                    'We will not continue to decompress.'
                )

        return resp

    # ------------------------------------------------------------------------------- #
    # Our hijacker request function
    # ------------------------------------------------------------------------------- #

    def request(self, method, url, *args, **kwargs):
        # pylint: disable=E0203
        if kwargs.get('proxies') and kwargs.get('proxies') != self.proxies:
            self.proxies = kwargs.get('proxies')

        # ------------------------------------------------------------------------------- #
        # Pre-Hook the request via user defined function.
        # ------------------------------------------------------------------------------- #

        if self.requestPreHook:
            (method, url, args, kwargs) = self.requestPreHook(
                self,
                method,
                url,
                *args,
                **kwargs
            )

        # ------------------------------------------------------------------------------- #
        # Make the request via requests.
        # ------------------------------------------------------------------------------- #

        response = self.decodeBrotli(
            self.perform_request(method, url, *args, **kwargs)
        )

        # ------------------------------------------------------------------------------- #
        # Debug the request via the Response object.
        # ------------------------------------------------------------------------------- #

        if self.debug:
            self.debugRequest(response)

        # ------------------------------------------------------------------------------- #
        # Post-Hook the request aka Post-Hook the response via user defined function.
        # ------------------------------------------------------------------------------- #

        if self.requestPostHook:
            response = self.requestPostHook(self, response)

            if self.debug:
                self.debugRequest(response)

        # Check if Cloudflare anti-bot is on
        if self.is_Challenge_Request(response):
            # ------------------------------------------------------------------------------- #
            # Try to solve the challenge and send it back
            # ------------------------------------------------------------------------------- #

            if self._solveDepthCnt >= self.solveDepth:
                _ = self._solveDepthCnt
                self.simpleException(
                    CloudflareLoopProtection,
                    f"!!Loop Protection!! We have tried to solve {_} time(s) in a row."
                )

            self._solveDepthCnt += 1

            response = self.Challenge_Response(response, **kwargs)
        else:
            if not response.is_redirect and response.status_code not in [429, 503]:
                self._solveDepthCnt = 0

        return response

    # ------------------------------------------------------------------------------- #
    # check if the response contains a valid Cloudflare Bot Fight Mode challenge
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def is_BFM_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and re.search(
                    r"\/cdn-cgi\/bm\/cv\/\d+\/api\.js.*?"
                    r"window\['__CF\$cv\$params'\]\s*=\s*{",
                    resp.text,
                    re.M | re.S
                )
            )
        except AttributeError:
            pass

        return False

    # ------------------------------------------------------------------------------- #
    # check if the response contains a valid Cloudflare challenge
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def is_IUAM_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code in [429, 503]
                and re.search(
                    r'<form .*?="challenge-form" action="/.*?__cf_chl_jschl_tk__=\S+"',
                    resp.text,
                    re.M | re.S
                )
            )
        except AttributeError:
            pass

        return False

    # ------------------------------------------------------------------------------- #
    # check if the response contains new Cloudflare challenge
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def is_New_IUAM_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code in [429, 503]
                and re.search(
                    r'cpo.src\s*=\s*"/cdn-cgi/challenge-platform/\S+orchestrate/jsch/v1"',
                    resp.text,
                    re.M | re.S
                )
                and re.search(r'window._cf_chl_enter\s*[\(=]', resp.text, re.M | re.S)
            )
        except AttributeError:
            pass

        return False

    # ------------------------------------------------------------------------------- #
    # check if the response contains a v2 hCaptcha Cloudflare challenge
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def is_New_Captcha_Challenge(resp):
        try:
            return (
                CloudScraper.is_Captcha_Challenge(resp)
                and re.search(
                    r'cpo.src\s*=\s*"/cdn-cgi/challenge-platform/\S+orchestrate/captcha/v1"',
                    resp.text,
                    re.M | re.S
                )
                and re.search(r'\s*id="trk_captcha_js"', resp.text, re.M | re.S)
            )
        except AttributeError:
            pass

        return False

    # ------------------------------------------------------------------------------- #
    # check if the response contains a Cloudflare hCaptcha challenge
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def is_Captcha_Challenge(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code == 403
                and re.search(
                    r'action="/\S+__cf_chl_captcha_tk__=\S+',
                    resp.text,
                    re.M | re.DOTALL
                )
            )
        except AttributeError:
            pass

        return False

    # ------------------------------------------------------------------------------- #
    # check if the response contains Firewall 1020 Error
    # ------------------------------------------------------------------------------- #

    @staticmethod
    def is_Firewall_Blocked(resp):
        try:
            return (
                resp.headers.get('Server', '').startswith('cloudflare')
                and resp.status_code == 403
                and re.search(
                    r'<span class="cf-error-code">1020</span>',
                    resp.text,
                    re.M | re.DOTALL
                )
            )
        except AttributeError:
            pass

        return False

    # ------------------------------------------------------------------------------- #
    # Wrapper for is_Captcha_Challenge, is_IUAM_Challenge, is_Firewall_Blocked
    # ------------------------------------------------------------------------------- #

    def is_Challenge_Request(self, resp):
        if self.is_Firewall_Blocked(resp):
            self.simpleException(
                CloudflareCode1020,
                'Cloudflare has blocked this request (Code 1020 Detected).'
            )

        if self.is_New_Captcha_Challenge(resp):
            self.simpleException(
                CloudflareChallengeError,
                'Detected a Cloudflare version 2 Captcha challenge, This feature is not available in the opensource (free) version.'
            )

        if self.is_New_IUAM_Challenge(resp):
            self.simpleException(
                CloudflareChallengeError,
                'Detected a Cloudflare version 2 challenge, This feature is not available in the opensource (free) version.'
            )

        if self.is_Captcha_Challenge(resp) or self.is_IUAM_Challenge(resp):
            if self.debug:
                print('Detected a Cloudflare version 1 challenge.')
            return True

        return False

    # ------------------------------------------------------------------------------- #
    # Try to solve cloudflare javascript challenge.
    # ------------------------------------------------------------------------------- #

    def IUAM_Challenge_Response(self, body, url, interpreter):
        try:
            formPayload = re.search(
                r'<form (?P<form>.*?="challenge-form" '
                r'action="(?P<challengeUUID>.*?'
                r'__cf_chl_jschl_tk__=\S+)"(.*?)</form>)',
                body,
                re.M | re.DOTALL
            ).groupdict()

            if not all(key in formPayload for key in ['form', 'challengeUUID']):
                self.simpleException(
                    CloudflareIUAMError,
                    "Cloudflare IUAM detected, unfortunately we can't extract the parameters correctly."
                )

            payload = OrderedDict()
            for challengeParam in re.findall(r'^\s*<input\s(.*?)/>', formPayload['form'], re.M | re.S):
                inputPayload = dict(re.findall(r'(\S+)="(\S+)"', challengeParam))
                if inputPayload.get('name') in ['r', 'jschl_vc', 'pass']:
                    payload.update({inputPayload['name']: inputPayload['value']})

        except AttributeError:
            self.simpleException(
                CloudflareIUAMError,
                "Cloudflare IUAM detected, unfortunately we can't extract the parameters correctly."
            )

        hostParsed = urlparse(url)

        try:
            payload['jschl_answer'] = JavaScriptInterpreter.dynamicImport(
                interpreter
            ).solveChallenge(body, hostParsed.netloc)
        except Exception as e:
            self.simpleException(
                CloudflareIUAMError,
                f"Unable to parse Cloudflare anti-bots page: {getattr(e, 'message', e)}"
            )

        return {
            'url': f"{hostParsed.scheme}://{hostParsed.netloc}{self.unescape(formPayload['challengeUUID'])}",
            'data': payload
        }

    # ------------------------------------------------------------------------------- #
    #  Try to solve the Captcha challenge via 3rd party.
    # ------------------------------------------------------------------------------- #

    def captcha_Challenge_Response(self, provider, provider_params, body, url):
        try:
            formPayload = re.search(
                r'<form (?P<form>.*?="challenge-form" '
                r'action="(?P<challengeUUID>.*?__cf_chl_captcha_tk__=\S+)"(.*?)</form>)',
                body,
                re.M | re.DOTALL
            ).groupdict()

            if not all(key in formPayload for key in ['form', 'challengeUUID']):
                self.simpleException(
                    CloudflareCaptchaError,
                    "Cloudflare Captcha detected, unfortunately we can't extract the parameters correctly."
                )

            payload = OrderedDict(
                re.findall(
                    r'(name="r"\svalue|data-ray|data-sitekey|name="cf_captcha_kind"\svalue)="(.*?)"',
                    formPayload['form']
                )
            )

            captchaType = 'reCaptcha' if payload['name="cf_captcha_kind" value'] == 're' else 'hCaptcha'

        except (AttributeError, KeyError):
            self.simpleException(
                CloudflareCaptchaError,
                "Cloudflare Captcha detected, unfortunately we can't extract the parameters correctly."
            )

        # ------------------------------------------------------------------------------- #
        # Pass proxy parameter to provider to solve captcha.
        # ------------------------------------------------------------------------------- #

        if self.proxies and self.proxies != self.captcha.get('proxy'):
            self.captcha['proxy'] = self.proxies

        # ------------------------------------------------------------------------------- #
        # Pass User-Agent if provider supports it to solve captcha.
        # ------------------------------------------------------------------------------- #

        self.captcha['User-Agent'] = self.headers['User-Agent']

        # ------------------------------------------------------------------------------- #
        # Submit job to provider to request captcha solve.
        # ------------------------------------------------------------------------------- #

        captchaResponse = Captcha.dynamicImport(
            provider.lower()
        ).solveCaptcha(
            captchaType,
            url,
            payload['data-sitekey'],
            provider_params
        )

        # ------------------------------------------------------------------------------- #
        # Parse and handle the response of solved captcha.
        # ------------------------------------------------------------------------------- #

        dataPayload = OrderedDict([
            ('r', payload.get('name="r" value', '')),
            ('cf_captcha_kind', payload['name="cf_captcha_kind" value']),
            ('id', payload.get('data-ray')),
            ('g-recaptcha-response', captchaResponse)
        ])

        if captchaType == 'hCaptcha':
            dataPayload.update({'h-captcha-response': captchaResponse})

        hostParsed = urlparse(url)

        return {
            'url': f"{hostParsed.scheme}://{hostParsed.netloc}{self.unescape(formPayload['challengeUUID'])}",
            'data': dataPayload
        }

    # ------------------------------------------------------------------------------- #
    # Attempt to handle and send the challenge response back to cloudflare
    # ------------------------------------------------------------------------------- #

    def Challenge_Response(self, resp, **kwargs):
        if self.is_Captcha_Challenge(resp):
            # ------------------------------------------------------------------------------- #
            # double down on the request as some websites are only checking
            # if cfuid is populated before issuing Captcha.
            # ------------------------------------------------------------------------------- #

            if self.doubleDown:
                resp = self.decodeBrotli(
                    self.perform_request(resp.request.method, resp.url, **kwargs)
                )

            if not self.is_Captcha_Challenge(resp):
                return resp

            # ------------------------------------------------------------------------------- #
            # if no captcha provider raise a runtime error.
            # ------------------------------------------------------------------------------- #

            if not self.captcha or not isinstance(self.captcha, dict) or not self.captcha.get('provider'):
                self.simpleException(
                    CloudflareCaptchaProvider,
                    "Cloudflare Captcha detected, unfortunately you haven't loaded an anti Captcha provider "
                    "correctly via the 'captcha' parameter."
                )

            # ------------------------------------------------------------------------------- #
            # if provider is return_response, return the response without doing anything.
            # ------------------------------------------------------------------------------- #

            if self.captcha.get('provider') == 'return_response':
                return resp

            # ------------------------------------------------------------------------------- #
            # Submit request to parser wrapper to solve captcha
            # ------------------------------------------------------------------------------- #

            submit_url = self.captcha_Challenge_Response(
                self.captcha.get('provider'),
                self.captcha,
                resp.text,
                resp.url
            )
        else:
            # ------------------------------------------------------------------------------- #
            # Cloudflare requires a delay before solving the challenge
            # ------------------------------------------------------------------------------- #

            if not self.delay:
                try:
                    delay = float(
                        re.search(
                            r'submit\(\);\r?\n\s*},\s*([0-9]+)',
                            resp.text
                        ).group(1)
                    ) / float(1000)
                    if isinstance(delay, (int, float)):
                        self.delay = delay
                except (AttributeError, ValueError):
                    self.simpleException(
                        CloudflareIUAMError,
                        "Cloudflare IUAM possibility malformed, issue extracing delay value."
                    )

            sleep(self.delay)

            # ------------------------------------------------------------------------------- #

            submit_url = self.IUAM_Challenge_Response(
                resp.text,
                resp.url,
                self.interpreter
            )

        # ------------------------------------------------------------------------------- #
        # Send the Challenge Response back to Cloudflare
        # ------------------------------------------------------------------------------- #

        if submit_url:

            def updateAttr(obj, name, newValue):
                try:
                    obj[name].update(newValue)
                    return obj[name]
                except (AttributeError, KeyError):
                    obj[name] = {}
                    obj[name].update(newValue)
                    return obj[name]

            cloudflare_kwargs = deepcopy(kwargs)
            cloudflare_kwargs['allow_redirects'] = False
            cloudflare_kwargs['data'] = updateAttr(
                cloudflare_kwargs,
                'data',
                submit_url['data']
            )

            urlParsed = urlparse(resp.url)
            cloudflare_kwargs['headers'] = updateAttr(
                cloudflare_kwargs,
                'headers',
                {
                    'Origin': f'{urlParsed.scheme}://{urlParsed.netloc}',
                    'Referer': resp.url
                }
            )

            challengeSubmitResponse = self.request(
                'POST',
                submit_url['url'],
                **cloudflare_kwargs
            )

            if challengeSubmitResponse.status_code == 400:
                self.simpleException(
                    CloudflareSolveError,
                    'Invalid challenge answer detected, Cloudflare broken?'
                )

            # ------------------------------------------------------------------------------- #
            # Return response if Cloudflare is doing content pass through instead of 3xx
            # else request with redirect URL also handle protocol scheme change http -> https
            # ------------------------------------------------------------------------------- #

            if not challengeSubmitResponse.is_redirect:
                return challengeSubmitResponse

            else:
                cloudflare_kwargs = deepcopy(kwargs)
                cloudflare_kwargs['headers'] = updateAttr(
                    cloudflare_kwargs,
                    'headers',
                    {'Referer': challengeSubmitResponse.url}
                )

                if not urlparse(challengeSubmitResponse.headers['Location']).netloc:
                    redirect_location = urljoin(
                        challengeSubmitResponse.url,
                        challengeSubmitResponse.headers['Location']
                    )
                else:
                    redirect_location = challengeSubmitResponse.headers['Location']

                return self.request(
                    resp.request.method,
                    redirect_location,
                    **cloudflare_kwargs
                )

        # ------------------------------------------------------------------------------- #
        # We shouldn't be here...
        # Re-request the original query and/or process again....
        # ------------------------------------------------------------------------------- #

        return self.request(resp.request.method, resp.url, **kwargs)

    # ------------------------------------------------------------------------------- #

    @classmethod
    def create_scraper(cls, sess=None, **kwargs):
        """
        Convenience function for creating a ready-to-go CloudScraper object.
        """
        scraper = cls(**kwargs)

        if sess:
            for attr in ['auth', 'cert', 'cookies', 'headers', 'hooks', 'params', 'proxies', 'data']:
                val = getattr(sess, attr, None)
                if val:
                    setattr(scraper, attr, val)

        return scraper

    # ------------------------------------------------------------------------------- #
    # Functions for integrating cloudscraper with other applications and scripts
    # ------------------------------------------------------------------------------- #

    @classmethod
    def get_tokens(cls, url, **kwargs):
        scraper = cls.create_scraper(
            **{
                field: kwargs.pop(field, None) for field in [
                    'allow_brotli',
                    'browser',
                    'debug',
                    'delay',
                    'doubleDown',
                    'captcha',
                    'interpreter',
                    'source_address'
                    'requestPreHook',
                    'requestPostHook'
                ] if field in kwargs
            }
        )

        try:
            resp = scraper.get(url, **kwargs)
            resp.raise_for_status()
        except Exception:
            logging.error(f'"{url}" returned an error. Could not collect tokens.')
            raise

        domain = urlparse(resp.url).netloc
        # noinspection PyUnusedLocal
        cookie_domain = None

        for d in scraper.cookies.list_domains():
            if d.startswith('.') and d in (f'.{domain}'):
                cookie_domain = d
                break
        else:
            cls.simpleException(
                CloudflareIUAMError,
                "Unable to find Cloudflare cookies. Does the site actually "
                "have Cloudflare IUAM (I'm Under Attack Mode) enabled?"
            )

        return (
            {
                '__cfduid': scraper.cookies.get('__cfduid', '', domain=cookie_domain),
                'cf_clearance': scraper.cookies.get('cf_clearance', '', domain=cookie_domain)
            },
            scraper.headers['User-Agent']
        )

    # ------------------------------------------------------------------------------- #

    @classmethod
    def get_cookie_string(cls, url, **kwargs):
        """
        Convenience function for building a Cookie HTTP header value.
        """
        tokens, user_agent = cls.get_tokens(url, **kwargs)
        return '; '.join('='.join(pair) for pair in tokens.items()), user_agent


# ------------------------------------------------------------------------------- #

if ssl.OPENSSL_VERSION_INFO < (1, 1, 1):
    print(
        f"DEPRECATION: The OpenSSL being used by this python install ({ssl.OPENSSL_VERSION}) does not meet the minimum supported "
        "version (>= OpenSSL 1.1.1) in order to support TLS 1.3 required by Cloudflare, "
        "You may encounter an unexpected Captcha or cloudflare 1020 blocks."
    )

# ------------------------------------------------------------------------------- #

create_scraper = CloudScraper.create_scraper
get_tokens = CloudScraper.get_tokens
get_cookie_string = CloudScraper.get_cookie_string



# coding: utf-8
"""
Sphinx directive integration
============================

We usually need to document the life-cycle of functions and classes:
when they are created, modified or deprecated.

To do that, `Sphinx <http://www.sphinx-doc.org>`_ has a set
of `Paragraph-level markups <http://www.sphinx-doc.org/en/stable/markup/para.html>`_:

- ``versionadded``: to document the version of the project which added the described feature to the library,
- ``versionchanged``: to document changes of a feature,
- ``deprecated``: to document a deprecated feature.

The purpose of this module is to defined decorators which adds this Sphinx directives
to the docstring of your function and classes.

Of course, the ``@deprecated`` decorator will emit a deprecation warning
when the function/method is called or the class is constructed.
"""
import textwrap

import wrapt

from deprecated.classic import ClassicAdapter
from deprecated.classic import deprecated as _classic_deprecated


class SphinxAdapter(ClassicAdapter):
    """
    Sphinx adapter -- *for advanced usage only*

    This adapter override the :class:`~deprecated.classic.ClassicAdapter`
    in order to add the Sphinx directives to the end of the function/class docstring.
    Such a directive is a `Paragraph-level markup <http://www.sphinx-doc.org/en/stable/markup/para.html>`_

    - The directive can be one of "versionadded", "versionchanged" or "deprecated".
    - The version number is added if provided.
    - The reason message is obviously added in the directive block if not empty.
    """

    def __init__(self, directive, reason="", version="", action=None, category=DeprecationWarning):
        """
        Construct a wrapper adapter.

        :type  directive: str
        :param directive:
            Sphinx directive: can be one of "versionadded", "versionchanged" or "deprecated".

        :type  reason: str
        :param reason:
            Reason message which documents the deprecation in your library (can be omitted).

        :type  version: str
        :param version:
            Version of your project which deprecates this feature.
            If you follow the `Semantic Versioning <https://semver.org/>`_,
            the version number has the format "MAJOR.MINOR.PATCH".

        :type  action: str
        :param action:
            A warning filter used to activate or not the deprecation warning.
            Can be one of "error", "ignore", "always", "default", "module", or "once".
            If ``None`` or empty, the the global filtering mechanism is used.
            See: `The Warnings Filter`_ in the Python documentation.

        :type  category: type
        :param category:
            The warning category to use for the deprecation warning.
            By default, the category class is :class:`~DeprecationWarning`,
            you can inherit this class to define your own deprecation warning category.
        """
        self.directive = directive
        super(SphinxAdapter, self).__init__(reason=reason, version=version, action=action, category=category)

    def __call__(self, wrapped):
        """
        Add the Sphinx directive to your class or function.

        :param wrapped: Wrapped class or function.

        :return: the decorated class or function.
        """
        reason = textwrap.dedent(self.reason).strip()
        reason = '\n'.join(
            textwrap.fill(line, width=70, initial_indent='   ', subsequent_indent='   ') for line in reason.splitlines()
        ).strip()
        docstring = textwrap.dedent(wrapped.__doc__ or "")
        if docstring:
            docstring += "\n\n"
        if self.version:
            docstring += ".. {directive}:: {version}\n".format(directive=self.directive, version=self.version)
        else:
            docstring += ".. {directive}::\n".format(directive=self.directive)
        if reason:
            docstring += "   {reason}\n".format(reason=reason)
        wrapped.__doc__ = docstring
        return super(SphinxAdapter, self).__call__(wrapped)


def versionadded(reason="", version=""):
    """
    This decorator can be used to insert a "versionadded" directive
    in your function/class docstring in order to documents the
    version of the project which adds this new functionality in your library.

    :param str reason:
        Reason message which documents the addition in your library (can be omitted).

    :param str version:
        Version of your project which adds this feature.
        If you follow the `Semantic Versioning <https://semver.org/>`_,
        the version number has the format "MAJOR.MINOR.PATCH", and,
        in the case of a new functionality, the "PATCH" component should be "0".

    :return: the decorated function.
    """
    adapter = SphinxAdapter('versionadded', reason=reason, version=version)

    # noinspection PyUnusedLocal
    @wrapt.decorator(adapter=adapter)
    def wrapper(wrapped, instance, args, kwargs):
        return wrapped(*args, **kwargs)

    return wrapper


def versionchanged(reason="", version=""):
    """
    This decorator can be used to insert a "versionchanged" directive
    in your function/class docstring in order to documents the
    version of the project which modifies this functionality in your library.

    :param str reason:
        Reason message which documents the modification in your library (can be omitted).

    :param str version:
        Version of your project which modifies this feature.
        If you follow the `Semantic Versioning <https://semver.org/>`_,
        the version number has the format "MAJOR.MINOR.PATCH".

    :return: the decorated function.
    """
    adapter = SphinxAdapter('versionchanged', reason=reason, version=version)

    # noinspection PyUnusedLocal
    @wrapt.decorator(adapter=adapter)
    def wrapper(wrapped, instance, args, kwargs):
        return wrapped(*args, **kwargs)

    return wrapper


def deprecated(*args, **kwargs):
    """
    This decorator can be used to insert a "deprecated" directive
    in your function/class docstring in order to documents the
    version of the project which deprecates this functionality in your library.

    Keyword arguments can be:

    -   "reason":
        Reason message which documents the deprecation in your library (can be omitted).

    -   "version":
        Version of your project which deprecates this feature.
        If you follow the `Semantic Versioning <https://semver.org/>`_,
        the version number has the format "MAJOR.MINOR.PATCH".

    -   "action":
        A warning filter used to activate or not the deprecation warning.
        Can be one of "error", "ignore", "always", "default", "module", or "once".
        If ``None``, empty or missing, the the global filtering mechanism is used.

    -   "category":
        The warning category to use for the deprecation warning.
        By default, the category class is :class:`~DeprecationWarning`,
        you can inherit this class to define your own deprecation warning category.

    :return: the decorated function.
    """
    directive = kwargs.pop('directive', 'deprecated')
    adapter_cls = kwargs.pop('adapter_cls', SphinxAdapter)
    return _classic_deprecated(*args, directive=directive, adapter_cls=adapter_cls, **kwargs)



import re
import traceback
from urllib.parse import urljoin

from requests.utils import dict_from_cookiejar

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("Speedcd")

        # Credentials
        self.username = None
        self.password = None

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0
        self.freeleech = False

        self.enable_cookies = True

        # URLs
        self.url = "https://speed.cd"
        self.urls = {
            "login": urljoin(self.url, "takeElogin.php"),
        }

        # Proper Strings
        self.proper_strings = ["PROPER", "REPACK"]

        # Cache
        self.cache = tvcache.TVCache(self)

    def login(self):
        if self.cookies:
            logger.debug(self.add_cookies_from_ui()[1])

        if any(dict_from_cookiejar(self.session.cookies).values()):
            return True

        login_params = {
            "username": self.username,
            "password": self.password,
        }

        # Yay lets add another request to the process since they are unreasonable.
        response = self.get_url(self.url, returns="text")
        with BS4Parser(response, "html5lib") as html:
            form = html.find("form", id="loginform")
            if form:
                self.urls["login"] = urljoin(self.url, form["action"])

        response = self.get_url(self.urls["login"], post_data=login_params, returns="text")
        if not response:
            logger.warning("Unable to connect to provider")
            return False

        if re.search("Incorrect username or Password. Please try again.", response):
            logger.warning("Invalid username or password. Check your settings")
            return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        results = []
        if not self.login():
            return results

        # http://speed.cd/browse/49/50/52/41/55/2/30/freeleech/deep/q/arrow
        # Search Params
        search_params = [
            "browse",
            "41",  # TV/Packs
            "2",  # Episodes
            "49",  # TV/HD
            "50",  # TV/Sports
            "52",  # TV/B-Ray
            "55",  # TV/Kids
            "30",  # Anime
        ]
        if self.freeleech:
            search_params.append("freeleech")
        search_params.append("deep")

        # Units
        units = ["B", "KB", "MB", "GB", "TB", "PB"]

        def process_column_header(td):
            result = ""
            img = td.find("img")
            if img:
                result = img.get("alt")
                if not result:
                    result = img.get("title")

            if not result:
                anchor = td.find("a")
                if anchor:
                    result = anchor.get_text(strip=True)

            if not result:
                result = td.get_text(strip=True)
            return result

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in {*search_strings[mode]}:
                current_params = search_params
                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))
                    current_params += ["q", re.sub(r"[^\w\s]", "", search_string)]

                data = self.get_url(urljoin(self.url, "/".join(current_params)), returns="text")
                if not data:
                    continue

                with BS4Parser(data, "html5lib") as html:
                    torrent_table = html.find("div", class_="boxContent")
                    torrent_table = torrent_table.find("table") if torrent_table else []
                    # noinspection PyCallingNonCallable
                    torrent_rows = torrent_table("tr") if torrent_table else []

                    # Continue only if at least one Release is found
                    if len(torrent_rows) < 2:
                        logger.debug("Data returned from provider does not contain any torrents")
                        continue

                    labels = [process_column_header(label) for label in torrent_rows[0]("th")]
                    row_labels = [process_column_header(label) for label in torrent_rows[1]("td")]

                    def label_index(name):
                        if name in labels:
                            return labels.index(name)
                        return row_labels.index(name)

                    # Skip column headers
                    for result in torrent_rows[1:]:
                        try:
                            cells = result("td")
                            title = cells[label_index("Title")].find("a").get_text()
                            download_url = urljoin(self.url, cells[label_index("Download")].a["href"])
                            if not all([title, download_url]):
                                continue

                            seeders = try_int(cells[label_index("Seeders") - 1].get_text(strip=True))
                            leechers = try_int(cells[label_index("Leechers") - 1].get_text(strip=True))

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            torrent_size = cells[label_index("Size") - 1].get_text()
                            size = convert_size(torrent_size, units=units) or -1

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug("Found result: {0} with {1} seeders and {2} leechers".format(title, seeders, leechers))

                            items.append(item)
                        except Exception as error:
                            logger.debug(f"Speed.cd: {error}")
                            logger.debug(traceback.format_exc())
                            continue

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



import re
from urllib.parse import urljoin

import validators

from sickchill import logger
from sickchill.helper.common import convert_size, try_int
from sickchill.oldbeard import tvcache
from sickchill.oldbeard.bs4_parser import BS4Parser
from sickchill.providers.torrent.TorrentProvider import TorrentProvider


class Provider(TorrentProvider):
    def __init__(self):

        # Provider Init
        super().__init__("YggTorrent")

        # Credentials
        self.username = None
        self.password = None

        # Torrent Stats
        self.minseed = 0
        self.minleech = 0

        # URLs
        self.custom_url = None
        self.url = "https://www2.yggtorrent.si/"
        self.urls = {"login": urljoin(self.url, "user/login"), "search": urljoin(self.url, "engine/search")}

        # Proper Strings
        self.proper_strings = ["PROPER"]

        # Cache
        self.cache = tvcache.TVCache(self, min_time=30)

    def update_urls(self, new_url, custom=False):
        if custom and not new_url:
            return True

        if not validators.url(new_url):
            if custom:
                logger.warning("Invalid custom url: {0}".format(self.custom_url))
            else:
                logger.debug("Url changing has failed!")

            return False

        self.url = new_url
        self.urls = {"login": urljoin(self.url, "user/login"), "search": urljoin(self.url, "engine/search")}
        return True

    def login(self):
        login_params = {
            "id": self.username,
            "pass": self.password,
        }

        self.update_urls(self.custom_url, True)

        response = self.get_url(self.urls["login"], post_data=login_params, returns="response")
        if response and self.url not in response.url:
            new_url = response.url.split("user/login")[0]
            logger.debug("Changing base url from {} to {}".format(self.url, new_url))
            if not self.update_urls(new_url):
                return False

            response = self.get_url(self.urls["login"], post_data=login_params, returns="response")

        # The login is now an AJAX call (401 : Bad credentials, 200 : Logged in, other : server failure)
        if not response or response.status_code != 200:
            logger.warning("Unable to connect to provider")
            return False
        else:
            # It seems we are logged, let's verify that !
            response = self.get_url(self.url, returns="response")

            if response.status_code != 200:
                logger.warning("Unable to connect to provider")
                return False
            if "logout" not in response.text:
                logger.warning("Invalid username or password. Check your settings")
                return False

        return True

    def search(self, search_strings, age=0, ep_obj=None):
        self.login()

        results = []

        for mode in search_strings:
            items = []
            logger.debug(_("Search Mode: {mode}".format(mode=mode)))

            for search_string in {*search_strings[mode]}:

                if mode != "RSS":
                    logger.debug(_("Search String: {search_string}".format(search_string=search_string)))
                # search string needs to be normalized, single quotes are apparently not allowed on the site
                # ç should also be replaced, people tend to use c instead
                replace_chars = {"'": "", "ç": "c"}

                for k, v in replace_chars.items():
                    search_string = search_string.replace(k, v)

                logger.debug("Sanitized string: {0}".format(search_string))

                try:
                    search_params = {"category": "2145", "sub_category": "all", "name": re.sub(r"[()]", "", search_string), "do": "search"}

                    data = self.get_url(self.urls["search"], params=search_params, returns="text")
                    if not data:
                        continue

                    if "logout" not in data:
                        logger.debug("Refreshing cookies")
                        self.login()

                    with BS4Parser(data, "html5lib") as html:
                        torrent_table = html.find(class_="table")
                        torrent_rows = torrent_table("tr") if torrent_table else []

                        # Continue only if at least one Release is found
                        if len(torrent_rows) < 2:
                            logger.debug("Data returned from provider does not contain any torrents")
                            continue

                        # Skip column headers
                        for result in torrent_rows[1:]:
                            cells = result("td")
                            if len(cells) < 9:
                                continue

                            title = cells[1].find("a").get_text(strip=True)
                            id = cells[2].find("a")["target"]
                            download_url = urljoin(self.url, "engine/download_torrent?id=" + id)

                            if not (title and download_url):
                                continue

                            seeders = try_int(cells[7].get_text(strip=True))
                            leechers = try_int(cells[8].get_text(strip=True))

                            torrent_size = cells[5].get_text()
                            size = convert_size(torrent_size) or -1

                            # Filter unseeded torrent
                            if seeders < self.minseed or leechers < self.minleech:
                                if mode != "RSS":
                                    logger.debug(
                                        "Discarding torrent because it doesn't meet the minimum seeders or leechers: {0} (S:{1} L:{2})".format(
                                            title, seeders, leechers
                                        )
                                    )
                                continue

                            item = {"title": title, "link": download_url, "size": size, "seeders": seeders, "leechers": leechers, "hash": ""}
                            if mode != "RSS":
                                logger.debug(
                                    _(
                                        "Found result: {title} with {seeders} seeders and {leechers} leechers".format(
                                            title=title, seeders=seeders, leechers=leechers
                                        )
                                    )
                                )

                            items.append(item)

                except (AttributeError, TypeError, KeyError, ValueError):
                    logger.exception("Failed parsing provider {}.".format(self.name))

            # For each search mode sort all the items by seeders if available
            items.sort(key=lambda d: try_int(d.get("seeders", 0)), reverse=True)
            results += items

        return results



"""
Exception classes
"""


class Error(Exception):
    """
    Pysrt's base exception
    """
    pass


class InvalidTimeString(Error):
    """
    Raised when parser fail on bad formated time strings
    """
    pass


class InvalidItem(Error):
    """
    Raised when parser fail to parse a sub title item
    """
    pass


class InvalidIndex(InvalidItem):
    """
    Raised when parser fail to parse a sub title index
    """
    pass



from .method import RPCMethod


class RPCCall(object):
    def __init__(self, rpc_method: RPCMethod, *args):
        self.rpc_method = rpc_method
        self.args = list(args)

    def get_method(self) -> RPCMethod:
        return self.rpc_method

    def get_args(self) -> list:
        return self.args

    def do_pre_processing(self):
        for processor in self.get_method().get_pre_processors():
            self.args = processor(*self.get_args())

    def do_post_processing(self, result):
        for processor in self.get_method().get_post_processors():
            result = processor(result)

        return result


import traceback

from tornado.web import RequestHandler

from sickchill import logger, settings
from sickchill.oldbeard import helpers


class KeyHandler(RequestHandler):
    def data_received(self, chunk):
        pass

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get(self, *args, **kwargs):
        if self.get_query_argument("u", None) == settings.WEB_USERNAME and self.get_query_argument("p", None) == settings.WEB_PASSWORD:
            if not len(settings.API_KEY or ""):
                settings.API_KEY = helpers.generateApiKey()
            result = {"success": True, "api_key": settings.API_KEY}
        else:
            result = {"success": False, "error": _("Failed authentication while getting api key")}
            logger.warning(_("Authentication failed during api key request: {traceback}".format(traceback=traceback.format_exc())))

        return self.finish(result)



from unittest import TestCase
from fanart.core import Request
from fanart.errors import RequestFanartError, ResponseFanartError
from httpretty import httprettified, HTTPretty


class RequestTestCase(TestCase):
    def test_valitate_error(self):
        self.assertRaises(RequestFanartError, Request, 'key', 'id', 'sport')

    @httprettified
    def test_response_error(self):
        request = Request('apikey', 'objid', 'tv')
        HTTPretty.register_uri(
            HTTPretty.GET,
            'http://webservice.fanart.tv/v3/tv/objid?api_key=apikey',
            body='Please specify a valid API key',
        )
        try:
            request.response()
        except ResponseFanartError as e:
            self.assertEqual(repr(e), "ResponseFanartError('Expecting value: "
                                      "line 1 column 1 (char 0)',)")
            self.assertEqual(str(e), "Expecting value: "
                                     "line 1 column 1 (char 0)")
